# Copyright (c) 2025 Marcin Zdun
# This code is licensed under MIT license (see LICENSE for details)

"""
The **proj_flow.project.cplusplus** registers a ``"C++"`` projects support.
"""

from proj_flow.project.cplusplus import cmake_context, conan_context

__all__ = ["cmake_context", "conan_context"]



# Copyright (c) 2026 Marcin Zdun
# This code is licensed under MIT license (see LICENSE for details)

import os
import random
import string
from dataclasses import dataclass, field, replace

from proj_flow.ext.test_runner.driver.test import Env, Test, fix_file_write


class color:
    reset = "\033[m"
    counter = "\033[2;49;92m"
    name = "\033[0;49;90m"
    failed = "\033[0;49;91m"
    passed = "\033[2;49;92m"
    skipped = "\033[0;49;34m"


class TaskResult:
    OK = 0
    SKIPPED = 1
    SAVED = 2
    FAILED = 3
    CLIP_FAILED = 4


@dataclass
class Counters:
    error_counter: int = 0
    skip_counter: int = 0
    save_counter: int = 0
    echo: list[str] = field(default_factory=list)

    def report(self, outcome: int, test_id: str, message: str | None):
        if outcome == TaskResult.SKIPPED:
            print(f"{test_id} {color.skipped}SKIPPED{color.reset}")
            self.skip_counter += 1
            return

        if outcome == TaskResult.SAVED:
            print(f"{test_id} {color.skipped}saved{color.reset}")
            self.skip_counter += 1
            self.save_counter += 1
            return

        if outcome == TaskResult.CLIP_FAILED:
            msg = f"{test_id} {color.failed}FAILED (unknown check '{message}'){color.reset}"
            print(msg)
            self.echo.append(msg)
            self.error_counter += 1
            return

        if outcome == TaskResult.OK:
            print(f"{test_id} {color.passed}PASSED{color.reset}")
            return

        if message is not None:
            print(message)
        msg = f"{test_id} {color.failed}FAILED{color.reset}"
        print(msg)
        self.echo.append(msg)
        self.error_counter += 1

    def summary(self, counter: int):
        print(f"Failed {self.error_counter}/{counter}")
        if self.skip_counter > 0:
            skip_test = "test" if self.skip_counter == 1 else "tests"
            if self.save_counter > 0:
                print(
                    f"Skipped {self.skip_counter} {skip_test} (including {self.save_counter} due to saving)"
                )
            else:
                print(f"Skipped {self.skip_counter} {skip_test}")

        if len(self.echo):
            print()
        for echo in self.echo:
            print(echo)

        return self.error_counter == 0


def task(
    env1: Env, tested: Test, current_counter: int
) -> tuple[int, str, str | None, str]:
    temp_instance = "".join(random.choice(string.ascii_letters) for _ in range(16))
    tempdir = f"{env1.tempdir}/{temp_instance}"
    tempdir_alt = None

    if env1.tempdir_alt is not None:
        tempdir_alt = f"{env1.tempdir_alt}{os.sep}{temp_instance}"

    env2 = replace(env1, tempdir=tempdir, tempdir_alt=tempdir_alt)

    test_counter = f"{color.counter}[{current_counter:>{env2.counter_digits}}/{env2.counter_total}]{color.reset}"
    test_name = f"{color.name}{tested.name}{color.reset}"
    test_id = f"{test_counter} {test_name}"

    print(test_id)
    os.makedirs(tempdir, exist_ok=True)

    result = tested.run(env2)
    if result is None:
        return (TaskResult.SKIPPED, test_id, None, tempdir)

    actual, files = result
    reports: list[str] = []

    saved = False
    if tested.expected is None:
        tested.data["expected"] = actual.as_dict()
        tested.store()
        saved = True

    for file in files:
        fixed = fix_file_write(file, env2, tested.cwd, tested.patches)

        if fixed.needs_saving():
            copied = fixed.copy_file()
            saved = saved or copied

        elif fixed.generated.content != fixed.template.content:
            reports.append(tested.report_file(fixed))

    if saved:
        return (TaskResult.SAVED, test_id, None, tempdir)

    clipped = tested.clip(actual)

    if isinstance(clipped, str):
        return (TaskResult.CLIP_FAILED, test_id, clipped, tempdir)

    if actual != tested.expected and clipped != tested.expected:
        reports.append(tested.report_io(actual))

    if reports:
        reports.append(tested.test_footer(env2, tempdir))
        return (TaskResult.FAILED, test_id, "\n".join(reports), tempdir)

    return (TaskResult.OK, test_id, None, tempdir)



# Copyright (c) 2026 Marcin Zdun
# This code is licensed under MIT license (see LICENSE for details)

from proj_flow.ext.webidl.cli.cmake import cmake
from proj_flow.ext.webidl.cli.depfile import depfile
from proj_flow.ext.webidl.cli.gen import gen
from proj_flow.ext.webidl.cli.init import init
from proj_flow.ext.webidl.cli.root import webidl

__all__ = ["webidl", "init", "cmake", "depfile", "gen"]



# Copyright (c) 2026 Marcin Zdun
# This code is licensed under MIT license (see LICENSE for details)


from dataclasses import dataclass, field
from pathlib import Path
from typing import Any, TypeVar, cast

from pywebidl2 import expr, parser, raw_parse, validate

from proj_flow.ext.webidl.base.config import ExtAttrsContextBuilders, TypeReplacement


def _flatten_union(ast: expr.IdlType) -> expr.IdlType:
    arguments = cast(list[expr.IdlType], ast.idl_type)
    result: list[expr.IdlType] = []
    for argument in arguments:
        if not argument.union:
            result.append(argument)
            continue
        argument = _flatten_union(argument)
        if argument.nullable:
            result.append(argument)
        sub_args = cast(list[expr.IdlType], argument.idl_type)
        result.extend(sub_args)
    ast.idl_type = result
    return ast


@dataclass
class TypeVisitor:
    known_types: set[str]
    types: dict[str, TypeReplacement]
    local_references: set[str] = field(default_factory=set)
    modules_or_includes: set[str] = field(default_factory=set)

    def on_type_name(self, type_name: str):
        if type_name in self.known_types:
            self.local_references.add(type_name)
            return type_name

        replacement = self.types.get(type_name)
        if replacement:
            if replacement.module_or_include:
                self.modules_or_includes.add(replacement.module_or_include)
            return replacement.name

        return type_name

    def on_type(self, type: "Type"):
        if type.nullable:
            replacement = self.types.get("optional")
            if replacement and replacement.module_or_include:
                self.modules_or_includes.add(replacement.module_or_include)

        type.name = self.on_type_name(type.idl_name)


@dataclass
class TypeArg:
    arg_type: "Type | expr.IdlType"
    first: bool = False


@dataclass
class Type:
    idl_name: str
    name: str | None = field(default=None)
    arguments: list[TypeArg] = field(default_factory=list)
    nullable: bool = field(default=False)
    generic: bool = field(default=False)
    ext_attrs: dict = field(default_factory=dict)

    def populate_references(self, visitor: TypeVisitor):
        visitor.on_type(self)
        for arg in self.arguments:
            if isinstance(arg.arg_type, Type):
                arg.arg_type.populate_references(visitor)

    def __patched_type(self):
        if self.generic and self.idl_name == "sequence":
            if self.ext_attrs.get("unique"):
                self.idl_name = "set"
            elif self.ext_attrs.get("span"):
                self.idl_name = "span"
        elif self.generic and self.idl_name == "record":
            first_arg = (
                self.arguments[0].arg_type
                if self.arguments and isinstance(self.arguments[0].arg_type, Type)
                else None
            )
            if first_arg:
                first_arg.idl_name = self.ext_attrs.get("key") or "string"
        for attr in ["unique", "span", "key"]:
            if attr in self.ext_attrs:
                del self.ext_attrs[attr]
        return self

    @staticmethod
    def from_idl(
        ast: expr.IdlType,
        builders: ExtAttrsContextBuilders,
        parent_ast_attrs: list[expr.ExtendedAttribute],
    ) -> "Type | expr.IdlType":
        ext_attrs = builders.from_idl(lambda b: b.type, ast.ext_attrs)
        parent_attrs = builders.from_idl(lambda b: b.type, parent_ast_attrs)
        builders.property_fixup(ext_attrs, parent_attrs)
        nullable = ext_attrs.get("nullable", False)
        if "nullable" in ext_attrs:
            del ext_attrs["nullable"]

        if ast.union:
            ast = _flatten_union(ast)
            args = cast(list[expr.IdlType], ast.idl_type)
            arguments = [TypeArg(Type.from_idl(arg, builders, [])) for arg in args]
            if arguments:
                arguments[0].first = True
            return Type(
                idl_name="union",
                arguments=arguments,
                nullable=ast.nullable or nullable,
                generic=True,
                ext_attrs=ext_attrs,
            ).__patched_type()

        if isinstance(ast.idl_type, str):
            return Type(
                idl_name=ast.idl_type,
                nullable=ast.nullable or nullable,
                ext_attrs=ext_attrs,
            ).__patched_type()

        if ast.generic:
            arguments: list[TypeArg] = []
            if isinstance(ast.idl_type, str):
                arguments = [TypeArg(Type(idl_name=ast.idl_type), True)]
            elif isinstance(ast.idl_type, expr.IdlType):
                arguments = [TypeArg(Type.from_idl(ast.idl_type, builders, []), True)]
            else:
                arguments = [
                    TypeArg(Type.from_idl(arg, builders, []))
                    for arg in cast(list[expr.IdlType], ast.idl_type)
                ]
                if arguments:
                    arguments[0].first = True
            return Type(
                idl_name=ast.generic,
                arguments=arguments,
                nullable=ast.nullable or nullable,
                generic=True,
                ext_attrs=ext_attrs,
            ).__patched_type()

        inner = ast.idl_type
        if isinstance(inner, expr.IdlType):
            inner.nullable = inner.nullable or ast.nullable or nullable
            inner.ext_attrs = [*ast.ext_attrs, *inner.ext_attrs]
            inner.generic = inner.generic or ast.generic
            return Type.from_idl(inner, builders, parent_ast_attrs)

        print(type(ast.idl_type), ast.generic != "")
        print(ast)
        return ast


@dataclass
class Attribute:
    name: str
    type: Type | expr.IdlType
    readonly: bool
    static: bool
    ext_attrs: dict

    def populate_references(self, visitor: TypeVisitor):
        if isinstance(self.type, Type):
            self.type.populate_references(visitor)

    @staticmethod
    def from_idl(ast: expr.Attribute, builders: ExtAttrsContextBuilders):
        ext_attrs = builders.from_idl(lambda b: b.attribute, ast.ext_attrs)
        type = Type.from_idl(ast.idl_type, builders, ast.ext_attrs)
        return Attribute(
            name=ast.name,
            type=type,
            readonly=ast.readonly,
            static=ast.special == "static",
            ext_attrs=ext_attrs,
        )


@dataclass
class Constant:
    name: str
    type: Type | expr.IdlType
    value: expr.Expression
    ext_attrs: dict

    def populate_references(self, visitor: TypeVisitor):
        if isinstance(self.type, Type):
            self.type.populate_references(visitor)

    @staticmethod
    def from_idl(ast: expr.Const, builders: ExtAttrsContextBuilders):
        ext_attrs = builders.from_idl(lambda b: b.attribute, ast.ext_attrs)
        type = Type.from_idl(ast.idl_type, builders, ast.ext_attrs)
        return Constant(
            name=ast.name,
            type=type,
            value=ast.value,
            ext_attrs=ext_attrs,
        )


@dataclass
class Argument:
    name: str
    type: Type | expr.IdlType
    ext_attrs: dict
    first: bool = False

    def populate_references(self, visitor: TypeVisitor):
        if isinstance(self.type, Type):
            self.type.populate_references(visitor)

    @staticmethod
    def from_idl(ast: expr.Argument, builders: ExtAttrsContextBuilders):
        ext_attrs = builders.from_idl(lambda b: b.argument, ast.ext_attrs)
        type = Type.from_idl(ast.idl_type, builders, ast.ext_attrs)
        return Argument(
            name=ast.name,
            type=type,
            ext_attrs=ext_attrs,
        )


@dataclass
class Operation:
    name: str
    type: Type | expr.IdlType | None
    arguments: list[Argument]
    static: bool
    ext_attrs: dict

    def populate_references(self, visitor: TypeVisitor):
        if isinstance(self.type, Type):
            self.type.populate_references(visitor)
        for arg in self.arguments:
            arg.populate_references(visitor)

    @staticmethod
    def from_idl(ast: expr.Operation, builders: ExtAttrsContextBuilders):
        ext_attrs = builders.from_idl(lambda b: b.operation, ast.ext_attrs)
        type = (
            Type.from_idl(ast.idl_type, builders, ast.ext_attrs)
            if ast.idl_type
            else None
        )
        arguments = [Argument.from_idl(arg, builders) for arg in ast.arguments]
        if arguments:
            arguments[0].first = True
        return Operation(
            name=ast.name,
            type=type,
            arguments=arguments,
            static=ast.special == "static",
            ext_attrs=ext_attrs,
        )


@dataclass
class MemberContainer:
    constants: list[Constant]
    attributes: list[Attribute]
    operations: list[Operation]

    @staticmethod
    def from_idl(
        ast: expr.Interface | expr.InterfaceMixin, builders: ExtAttrsContextBuilders
    ):
        constants: list[Constant] = []
        attributes: list[Attribute] = []
        operations: list[Operation] = []
        for member in ast.members:
            if isinstance(member, expr.Const):
                constants.append(Constant.from_idl(member, builders))
                continue

            if isinstance(member, expr.Attribute):
                attributes.append(Attribute.from_idl(member, builders))
                continue

            if isinstance(member, expr.Operation):
                operations.append(Operation.from_idl(member, builders))
                continue

            print(f"warning: unsupported member type `{getattr(member, 'type')}'")

        return MemberContainer(
            constants=constants,
            attributes=attributes,
            operations=operations,
        )


def update(existing: dict, incoming: dict):
    for key in incoming:
        val = incoming[key]
        if val or key not in existing:
            existing[key] = val


@dataclass
class Interface:
    name: str
    inheritance: str | None
    partial: bool
    constants: list[Constant]
    attributes: list[Attribute]
    operations: list[Operation]
    ext_attrs: dict
    has_constants: bool = field(default=False)
    has_attributes: bool = field(default=False)
    has_operations: bool = field(default=False)

    def populate_references(self, visitor: TypeVisitor):
        if self.inheritance:
            visitor.on_type_name(self.inheritance)
        for item in self.constants:
            item.populate_references(visitor)
        for item in self.attributes:
            item.populate_references(visitor)
        for item in self.operations:
            item.populate_references(visitor)

    @staticmethod
    def from_idl(ast: expr.Interface, builders: ExtAttrsContextBuilders):
        members = MemberContainer.from_idl(ast, builders)
        return Interface(
            name=ast.name,
            inheritance=ast.inheritance,
            partial=ast.partial,
            constants=members.constants,
            attributes=members.attributes,
            operations=members.operations,
            ext_attrs=builders.from_idl(lambda b: b.interface, ast.ext_attrs),
        )


@dataclass
class InterfaceMixin:
    partial: bool
    constants: list[Constant]
    attributes: list[Attribute]
    operations: list[Operation]
    ext_attrs: dict

    @staticmethod
    def from_idl(ast: expr.InterfaceMixin, builders: ExtAttrsContextBuilders):
        members = MemberContainer.from_idl(ast, builders)
        return InterfaceMixin(
            partial=ast.partial,
            constants=members.constants,
            attributes=members.attributes,
            operations=members.operations,
            ext_attrs=builders.from_idl(lambda b: b.interface, ast.ext_attrs),
        )


@dataclass
class EnumInfo:
    name: str
    NAME: str
    items: list[str]
    ext_attrs: dict

    @staticmethod
    def from_idl(ast: expr.Enum, builders: ExtAttrsContextBuilders):
        return EnumInfo(
            name=ast.name,
            NAME=ast.name.upper(),
            items=[cast(str, value.value) for value in ast.values],
            ext_attrs=builders.from_idl(lambda b: b.enum, ast.ext_attrs),
        )


@dataclass
class MergedDefinitions:
    interfaces: dict[str, Interface] = field(default_factory=dict)
    enum: list[EnumInfo] = field(default_factory=list)

    def order(self, types: dict[str, TypeReplacement]):
        dependency_tree: dict[str, set[str]] = {}
        modules_or_includes: set[str] = set()

        known_names = set[str]()
        known_names.update(item.name for item in self.enum)
        known_names.update(self.interfaces.keys())

        for name, interface in self.interfaces.items():
            visitor = TypeVisitor(known_names, types)
            interface.populate_references(visitor)
            modules_or_includes.update(visitor.modules_or_includes)
            if name in visitor.local_references:
                visitor.local_references.remove(name)
            dependency_tree[name] = visitor.local_references

        order = list[str]()
        while dependency_tree:
            layer = [
                name
                for name in sorted(dependency_tree.keys())
                if not dependency_tree[name]
            ]
            order.extend(layer)

            if not layer:
                for key in dependency_tree:
                    layer.append(key)
                    break

            for key in layer:
                del dependency_tree[key]

            for key in layer:
                for dependencies in dependency_tree.values():
                    if key in dependencies:
                        dependencies.remove(key)

        return ([self.interfaces[key] for key in order], modules_or_includes)


P_ = TypeVar("P_")
AST_ = TypeVar("AST_")


class PartialSetDict[P_, AST_](dict[str, list[P_]]):
    def append(self, name: str, item: AST_, builders: ExtAttrsContextBuilders):
        conv = self.convert(item, builders)
        try:
            self[name].append(conv)
        except KeyError:
            self[name] = [conv]

    def extend(self, additional: dict[str, list[P_]]):
        for name, partials in additional.items():
            try:
                self[name].extend(partials)
            except KeyError:
                self[name] = [*partials]

    def convert(self, item: AST_, builders: ExtAttrsContextBuilders) -> P_: ...


class InterfacePartialSet(PartialSetDict[Interface | str, expr.Interface | str]):
    def convert(self, item: expr.Interface | str, builders: ExtAttrsContextBuilders):
        return item if isinstance(item, str) else Interface.from_idl(item, builders)


class MixinPartialSet(PartialSetDict[InterfaceMixin, expr.InterfaceMixin]):
    def convert(self, item: expr.InterfaceMixin, builders: ExtAttrsContextBuilders):
        return InterfaceMixin.from_idl(item, builders)


@dataclass
class IdlSyntaxError:
    path: str
    error: parser.SyntaxErrorInfo


@dataclass
class Definitions:
    interfaces: InterfacePartialSet = field(default_factory=InterfacePartialSet)
    mixins: MixinPartialSet = field(default_factory=MixinPartialSet)
    enums: list[EnumInfo] = field(default_factory=list)

    @staticmethod
    def from_idl(ast: expr.Definitions, builders: ExtAttrsContextBuilders):
        result = Definitions()

        for definition in ast.definitions:
            if isinstance(definition, expr.Includes):
                result.interfaces.append(
                    definition.target, definition.includes, builders
                )
                continue
            if isinstance(definition, expr.Interface):
                result.interfaces.append(definition.name, definition, builders)
                continue
            if isinstance(definition, expr.InterfaceMixin):
                result.mixins.append(definition.name, definition, builders)
                continue
            if isinstance(definition, expr.Enum):
                result.enums.append(EnumInfo.from_idl(definition, builders))
                continue

        return result

    @staticmethod
    def merge(*partials: "Definitions"):
        intermediate = Definitions()
        for partial in partials:
            intermediate.interfaces.extend(partial.interfaces)
            intermediate.mixins.extend(partial.mixins)
            intermediate.enums.extend(partial.enums)

        interfaces: dict[str, Interface] = {}
        mixins: dict[str, Interface] = {}

        for name, mixin_set in intermediate.mixins.items():
            mixin = Interface(
                name=name,
                inheritance=None,
                partial=True,
                constants=[],
                attributes=[],
                operations=[],
                ext_attrs={},
            )

            for partial in mixin_set:
                mixin.constants.extend(partial.constants)
                mixin.attributes.extend(partial.attributes)
                mixin.operations.extend(partial.operations)
                update(mixin.ext_attrs, partial.ext_attrs)

            mixins[name] = mixin

        for name, interface_set in intermediate.interfaces.items():
            target = Interface(
                name=name,
                inheritance=None,
                partial=False,
                constants=[],
                attributes=[],
                operations=[],
                ext_attrs={},
            )

            for partial in interface_set:
                if isinstance(partial, str):
                    try:
                        partial = mixins[partial]
                    except KeyError:
                        message = f"{name} cannot find {partial} to include"
                        raise RuntimeError(message)

                target.inheritance = target.inheritance or partial.inheritance
                target.constants.extend(partial.constants)
                target.attributes.extend(partial.attributes)
                target.operations.extend(partial.operations)
                update(target.ext_attrs, partial.ext_attrs)

            target.has_constants = len(target.constants) > 0
            target.has_attributes = len(target.attributes) > 0
            target.has_operations = len(target.operations) > 0
            interfaces[name] = target

        return MergedDefinitions(
            interfaces=interfaces,
            enum=intermediate.enums,
        )

    @staticmethod
    def parse_and_merge(
        names: list[str],
        builders: ExtAttrsContextBuilders,
    ):
        partials: list[Definitions] = []
        syntax_errors: list[IdlSyntaxError] = []

        for name in names:
            path = Path(name)
            text = path.read_text(encoding="UTF-8")
            errors = validate(text)
            for error in errors:
                syntax_errors.append(IdlSyntaxError(path=name, error=error))
            if errors:
                continue

            ast = cast(expr.Definitions, raw_parse(text))
            definitions = Definitions.from_idl(ast, builders)
            partials.append(definitions)

        if syntax_errors:
            return syntax_errors
        return Definitions.merge(*partials)



# Copyright (c) 2024 Marcin Zdun
# This code is licensed under MIT license (see LICENSE for details)

"""
The **proj_flow.log.hosting** contains links information for various
hosting services. Currently, only GitHub is supported.
"""

from . import github

__all__ = ["github"]



# Copyright (c) 2026 Marcin Zdun
# This code is licensed under MIT license (see LICENSE for details)

import os
import random
import re
import shlex
import shutil
import string
import subprocess
import sys
from dataclasses import dataclass
from difflib import unified_diff
from pathlib import Path
from typing import Callable, TypeVar, cast

import yaml

try:
    from yaml import CDumper as Dumper
    from yaml import CLoader as Loader
except ImportError:
    from yaml import Dumper, Loader

_flds = ["Return code", "Standard out", "Standard err"]
_streams = ["stdout", "stderr"]


def _last_enter(value: str):
    if len(value) and value[-1] == "\n":
        value = value[:-1] + "\\n"
    return value + "\n"


def _diff(expected, actual):
    expected = _last_enter(expected).splitlines(keepends=False)
    actual = _last_enter(actual).splitlines(keepends=False)
    return "\n".join(list(unified_diff(expected, actual, lineterm=""))[2:])


def _alt_sep(input: str, value: str, var: str):
    split = input.split(value)
    first = split[0]
    split = split[1:]
    for index in range(len(split)):
        m = re.match(r"^(\S+)((\n|.)*)$", split[index])
        if m is None:
            continue
        g2 = m.group(2)
        if g2 is None:
            g2 = ""
        split[index] = "{}{}".format(m.group(1).replace(os.sep, "/"), g2)
    return var.join([first, *split])


def str_presenter(dumper: Dumper, data: str):
    if "\n" in data:
        return dumper.represent_scalar("tag:yaml.org,2002:str", data, style="|")

    return dumper.represent_scalar("tag:yaml.org,2002:str", data)


yaml.add_representer(str, str_presenter, Dumper=Dumper)


@dataclass
class Env:
    target: str
    target_name: str
    build_dir: str
    data_dir: str
    inst_dir: str
    tempdir: str
    version: str
    counter_digits: int
    counter_total: int
    handlers: dict[str, tuple[int, Callable[["Test", list[str]], None]]]
    data_dir_alt: str | None = None
    tempdir_alt: str | None = None
    # TODO: installable patches
    builtin_patches: dict[str, str] | None = None
    reportable_env_prefix: str | None = None

    def expand(
        self, input: str, tempdir: str, cwd: str, additional: dict[str, str] = {}
    ):
        input = (
            input.replace("$TMP", tempdir)
            .replace("$CWD", cwd)
            .replace("$DATA", self.data_dir)
            .replace("$INST", self.inst_dir)
            .replace("$VERSION", self.version)
        )
        for key, value in additional.items():
            input = input.replace(f"${key}", value)
        return input

    def fix(self, raw_input: bytes, cwd: str, patches: dict[str, str]):
        if os.name == "nt":
            raw_input = raw_input.replace(b"\r\n", b"\n")
        input = raw_input.decode("UTF-8")
        input = _alt_sep(input, cwd, "$CWD")
        input = _alt_sep(input, self.tempdir, "$TMP")
        input = _alt_sep(input, self.data_dir, "$DATA")
        input = input.replace(self.version, "$VERSION")

        if self.tempdir_alt is not None:
            input = _alt_sep(input, self.tempdir_alt, "$TMP")
            if self.data_dir_alt:
                input = _alt_sep(input, self.data_dir_alt, "$DATA")

        builtins = self.builtin_patches or {}

        lines = input.split("\n")
        for patch, replacement in builtins.items():
            pattern = re.compile(patch)
            for lineno in range(len(lines)):
                m = pattern.match(lines[lineno])
                if m:
                    lines[lineno] = m.expand(replacement)

        for patch, replacement in patches.items():
            pattern = re.compile(patch)
            for lineno in range(len(lines)):
                m = pattern.match(lines[lineno])
                if m:
                    lines[lineno] = m.expand(replacement)
        return "\n".join(lines)


def _test_name(filename: Path) -> str:
    dirname = filename.parent.name
    basename = filename.stem

    def num(s: str):
        items = s.split("-")
        if len(items) < 2:
            return s
        items[0] = f"({items[0]})"
        return " ".join(items)

    return f"{num(dirname)} :: {num(basename)}"


def _paths(key: str, dirs: list[str]):
    vals = [val for val in os.environ.get(key, "").split(os.pathsep) if val != ""]
    named = f"${key}"
    if named in dirs:
        pos = dirs.index(named)
        copy = [*dirs]
        copy[pos:pos] = vals
        vals = copy
    else:
        vals.extend(dirs)
    return os.pathsep.join(vals)


StrOrBytes = TypeVar("StrOrBytes", str, bytes)


@dataclass
class FileContents[StrOrBytes]:
    filename: str
    path: Path
    content: StrOrBytes | None

    def decode(self):
        return FileContents[str](
            filename=self.filename,
            path=self.path,
            content=cast(bytes, self.content).decode() if self.content else None,
        )


def load_file_contents(filename: str, environment: Env, cwd: str):
    path = Path(environment.expand(filename, environment.tempdir, cwd=cwd))
    try:
        content = path.read_bytes()
    except FileNotFoundError:
        content = None
    return FileContents(filename=filename, path=path, content=content)


def fix_file_contents(
    self: FileContents[bytes], env: Env, cwd: str, patches: dict[str, str]
):
    return FileContents(
        filename=self.filename,
        path=self.path,
        content=env.fix(self.content, cwd, patches) if self.content else None,
    )


@dataclass
class FileWrite[StrOrBytes]:
    generated: FileContents[StrOrBytes]
    template: FileContents[StrOrBytes]
    save: bool

    def needs_saving(self):
        return self.template.content is None or self.save

    def copy_file(self):
        if not self.generated.content:
            return False

        self.template.path.parent.mkdir(parents=True, exist_ok=True)
        self.template.path.write_bytes(cast(str, self.generated.content).encode())
        return True


def fix_file_write(self: FileWrite[bytes], env: Env, cwd: str, patches: dict[str, str]):
    return FileWrite(
        generated=fix_file_contents(self.generated, env, cwd, patches),
        template=self.template.decode(),
        save=self.save,
    )


@dataclass
class Expected:
    returncode: int
    stdout: str
    stderr: str

    def as_dict(self):
        return {
            "return-code": self.returncode,
            "stdout": self.stdout,
            "stderr": self.stderr,
        }


class Test:
    cwd: str
    data: dict
    filename: Path
    name: str
    ok: bool
    post_args: list[list[str]]
    current_env: Env | None
    additional_env: dict[str, str]

    linear: bool
    disabled: bool | str
    lang: str
    args: list[str]
    post: list[list[str]]
    expected: Expected | None
    check: list[str]
    writes: dict[str, str | dict]
    patches: dict[str, str]
    env: dict[str, str | None]
    prepare: list[list[str]]
    cleanup: list[list[str]]

    def __init__(self, data: dict, filename: Path, count: int):
        self.cwd = os.getcwd()
        self.data = data
        self.filename = filename
        self.name = _test_name(filename)
        self.ok = True
        self.post_args = []
        self.current_env = None
        self.additional_env = {}

        self.linear = cast(bool, data.get("linear", False))
        self.disabled = cast(bool | str, data.get("disabled", False))
        self.lang = cast(str, data.get("lang", "en"))
        self.args = []
        self.post = []
        self.expected = None
        self.check = ["all"] * len(_streams)
        self.writes = cast(dict[str, str | dict], data.get("writes", {}))
        self.patches = cast(dict[str, str], data.get("patches", {}))
        self.env = {}
        self.prepare = []
        self.cleanup = []

        if isinstance(self.disabled, bool):
            self.ok = not self.disabled
        elif isinstance(self.disabled, str):
            self.ok = self.disabled != sys.platform

        rebuild = False

        _args = cast(str | list[str] | None, data.get("args"))
        if _args is None:
            self.ok = False
            return
        elif isinstance(_args, list):
            rebuild = True
            data["args"] = shlex.join(_args)
            self.args = _args
        else:
            self.args = shlex.split(_args)

        self.post_args, rebuild = self.__join_args("post", data, rebuild)

        if "expected" not in data:
            self.ok = False
            return

        _expected = cast(None | dict[str, str | list[str]], data.get("expected"))
        if isinstance(_expected, dict):
            returncode = cast(None | int, _expected.get("return-code"))
            stdout = _expected.get("stdout")
            stderr = _expected.get("stderr")
            if not (
                isinstance(returncode, int)
                and isinstance(stdout, (str, list))
                and isinstance(stderr, (str, list))
            ):
                self.ok = False
                return
            if isinstance(stdout, list):
                stdout = "\n".join(stdout)
            if isinstance(stderr, list):
                stderr = "\n".join(stderr)
            self.expected = Expected(
                returncode=returncode, stdout=stdout, stderr=stderr
            )

        _check = cast(dict[str, str], data.get("check", {}))
        for index in range(len(_streams)):
            self.check[index] = _check.get(_streams[index], self.check[index])

        _env = cast(dict[str, str | list[str] | None], data.get("env", {}))
        self.env = {
            key: _paths(key, value) if isinstance(value, list) else value
            for key, value in _env.items()
        }

        self.prepare, rebuild = self.__join_args("prepare", data, rebuild)
        self.cleanup, rebuild = self.__join_args("cleanup", data, rebuild)

        if rebuild:
            if self.expected:
                data["expected"] = self.expected.as_dict()
            self.store()

    def __join_args(self, key: str, data: dict, rebuild: bool):
        _rebuild = rebuild
        result: list[list[str]] = []

        lines = cast(str | list[str | list[str]], data.get(key, []))
        if isinstance(lines, str):
            lines = [lines]

        for index in range(len(lines)):
            cmd = lines[index]
            if isinstance(cmd, str):
                result.append(shlex.split(cmd))
                continue

            _rebuild = True
            data[key][index] = shlex.join(cmd)
            result.append(cmd)

        return result, _rebuild

    def run_cmds(self, env: Env, ops: list[list[str]], tempdir: str) -> bool | None:
        saved = self.current_env
        self.current_env = env
        try:
            for op in ops:
                orig = [*op]
                is_safe = False
                try:
                    name = op[0]
                    if name[:5] == "safe-":
                        name = name[5:]
                        is_safe = True
                    min_args, cb = env.handlers[name]
                    op = op[1:]
                    if len(op) < min_args:
                        return None
                    cb(self, [env.expand(o, tempdir, cwd=self.cwd) for o in op])
                except Exception as ex:
                    if op[0] != "safe-rm":
                        print("Problem while handling", shlex.join(orig))
                        print(ex)
                        raise
                    if is_safe:
                        continue
                    return None
        finally:
            self.current_env = saved
        return True

    def run(self, environment: Env) -> tuple[Expected, list[FileWrite[bytes]]] | None:
        root = os.path.join(
            "build",
            ".testing",
            "".join(random.choice(string.ascii_letters) for _ in range(16)),
        )
        root = self.cwd = os.path.join(self.cwd, root)
        os.makedirs(root, exist_ok=True)

        prep = self.run_cmds(environment, self.prepare, environment.tempdir)
        if prep is None:
            return None

        expanded = [
            environment.expand(arg, environment.tempdir, self.cwd, self.additional_env)
            for arg in self.args
        ]
        post_expanded = [
            [
                environment.expand(
                    arg, environment.tempdir, self.cwd, self.additional_env
                )
                for arg in cmd
            ]
            for cmd in self.post_args
        ]

        _env = {name: os.environ[name] for name in os.environ}
        _env["LANGUAGE"] = self.lang
        for key in self.env:
            value = self.env[key]
            if value:
                _env[key] = environment.expand(value, environment.tempdir, cwd=self.cwd)
            elif key in _env:
                del _env[key]

        cwd = None if self.linear else self.cwd
        proc: subprocess.CompletedProcess = subprocess.run(
            [environment.target, *expanded], capture_output=True, env=_env, cwd=self.cwd
        )
        returncode: int = proc.returncode
        test_stdout: bytes = proc.stdout
        test_stderr: bytes = proc.stderr

        for sub_expanded in post_expanded:
            if returncode != 0:
                break
            proc_post: subprocess.CompletedProcess = subprocess.run(
                [environment.target, *sub_expanded],
                capture_output=True,
                env=_env,
                cwd=cwd,
            )
            returncode = proc_post.returncode
            if len(test_stdout) and len(proc_post.stdout):
                test_stdout += b"\n"
            if len(test_stderr) and len(proc_post.stderr):
                test_stderr += b"\n"
            test_stdout += proc_post.stdout
            test_stderr += proc_post.stderr

        expected_files: list[FileWrite[bytes]] = []
        for key, value in self.writes.items():
            if isinstance(value, str):
                expected_files.append(
                    FileWrite(
                        generated=load_file_contents(key, environment, cwd=self.cwd),
                        template=load_file_contents(value, environment, cwd=self.cwd),
                        save=False,
                    )
                )
            elif isinstance(value, dict):
                path = cast(str | None, value.get("path"))
                save = cast(bool, value.get("save", False))
                if not isinstance(path, str) or not isinstance(save, bool):
                    continue

                expected_files.append(
                    FileWrite(
                        generated=load_file_contents(key, environment, cwd=self.cwd),
                        template=load_file_contents(path, environment, cwd=self.cwd),
                        save=save,
                    )
                )

        clean = self.run_cmds(environment, self.cleanup, environment.tempdir)
        if clean is None:
            return None

        return (
            Expected(
                returncode=returncode,
                stdout=environment.fix(test_stdout, self.cwd, self.patches),
                stderr=environment.fix(test_stderr, self.cwd, self.patches),
            ),
            expected_files,
        )

    def clip(self, actual: Expected) -> str | Expected:
        if not self.expected:
            return actual

        returncode, streams = actual.returncode, [actual.stdout, actual.stderr]
        expected = [self.expected.stdout, self.expected.stderr]

        for ndx in range(len(self.check)):
            check = self.check[ndx]
            if check != "all":
                ex = expected[ndx]
                ex_len = len(ex)
                if check == "begin":
                    if len(streams[ndx]) > ex_len:
                        streams[ndx] = streams[ndx][:ex_len]
                elif check == "end":
                    streams[ndx] = streams[ndx][-ex_len:]
                else:
                    return check
        return Expected(returncode=returncode, stdout=streams[0], stderr=streams[1])

    @staticmethod
    def text_diff(
        header: str, expected: str, actual: str, pre_mark: str = "", post_mark: str = ""
    ):
        return f"""{header}
  Expected:
    {pre_mark}{repr(expected)}{post_mark}
  Actual:
    {pre_mark}{repr(actual)}{post_mark}

Diff:
{_diff(expected, actual)}
"""

    def report_io(self, actual: Expected):
        result = ""
        if not self.expected:
            return result

        if actual.returncode != self.expected.returncode:
            result += f"""{_flds[0]}
  Expected:
    {repr(self.expected.returncode)}
  Actual:
    {repr(actual.returncode)}
"""
        streams = [
            (_flds[1], self.check[0], actual.stdout, self.expected.stdout),
            (_flds[2], self.check[1], actual.stderr, self.expected.stderr),
        ]
        for header, check, actual_stream, expected_stream in streams:
            if actual_stream == expected_stream:
                continue

            if result:
                result += "\n"

            pre_mark = "..." if check == "end" else ""
            post_mark = "..." if check == "begin" else ""
            result += Test.text_diff(
                header,
                expected_stream,
                actual_stream,
                pre_mark,
                post_mark,
            )

        return result

    def report_file(
        self,
        file: FileWrite[str],
    ):
        header = file.generated.filename

        if not file.generated.content:
            header += "\n  New file was not present to be read"
        if not file.generated.content:
            header += (
                f"\n  Test file {file.generated.filename} was not present to be read"
            )

        if not file.generated.content or not file.template.content:
            return header

        return Test.text_diff(
            header,
            expected=file.template.content,
            actual=file.generated.content,
        )

    def test_footer(self, env: Env, tempdir: str):
        _env = {}
        _env["LANGUAGE"] = self.lang
        for key in self.env:
            value = self.env[key]
            if value:
                _env[key] = value
            elif key in _env:
                del _env[key]
        if env.reportable_env_prefix:
            for key in os.environ:
                if key.startswith(env.reportable_env_prefix) and key not in _env:
                    _env[key] = os.environ[key]

        expanded = [env.expand(arg, tempdir, cwd=self.cwd) for arg in self.args]
        call = " ".join(
            shlex.quote(arg.replace(os.sep, "/"))
            for arg in [
                *["{}={}".format(key, val) for key, val in _env.items()],
                env.target,
                *expanded,
            ]
        )
        return f"{call}\ncwd: {self.cwd}\ntest: {self.filename}"

    def nullify(self, lang: str | None):
        if lang is not None:
            self.lang = lang
            self.data["lang"] = lang
        self.expected = None
        self.data["expected"] = None
        self.store()

    def store(self):
        with self.filename.open("wb") as f:
            yaml.dump(
                self.data,
                stream=f,
                Dumper=Dumper,
                width=1024,
                allow_unicode=True,
                default_flow_style=False,
                sort_keys=False,
                encoding="UTF-8",
            )

    def path(self, filename):
        return os.path.join(self.cwd, filename)

    def chdir(self, sub):
        self.cwd = os.path.abspath(os.path.join(self.cwd, sub))
        if self.linear:
            os.chdir(self.cwd)

    def ls(self, sub):
        name = self.path(sub)
        for _, dirnames, filenames in os.walk(name):
            names = sorted(
                [
                    *((name.lower(), f"{name}/") for name in dirnames),
                    *((name.lower(), f"{name}") for name in filenames),
                ]
            )
            dirnames[:] = []
            for _, name in names:
                print(name)

    def rmtree(self, sub):
        shutil.rmtree(self.path(sub))

    def cp(self, src: str, dst: str):
        shutil.copy2(self.path(src), self.path(dst))

    def makedirs(self, sub):
        os.makedirs(self.path(sub), exist_ok=True)

    def store_output(self, name: str, args: list[str]):
        env = self.current_env

        if env is not None and args[0] == env.target_name:
            args[0] = env.target

        proc = subprocess.run(args, shell=False, capture_output=True, cwd=self.cwd)
        self.additional_env[name] = proc.stdout.decode("UTF-8").strip()
        print(f"export {name}={self.additional_env[name]}")

    @staticmethod
    def load(filename: Path, count: int):
        blob = filename.read_bytes().replace(b"\r\n", b"\n")
        tree = cast(dict, yaml.load(blob, Loader=Loader))
        return Test(tree, filename, count)



# Copyright (c) 2025 Marcin Zdun
# This code is licensed under MIT license (see LICENSE for details)

"""
The **proj_flow.project.cplusplus** registers a ``"C++"`` projects support.
"""

from proj_flow.base import path_get
from proj_flow.project import api

CPP_EXTENSIONS = [
    "proj_flow.ext.cplusplus",
    "proj_flow.ext.sign",
    "proj_flow.ext.store",
]


@api.project_type.add
class CPlusPlus(api.ProjectType):
    def __init__(self):
        super().__init__("C++ plus CMake plus Conan", "cxx")

    def get_extension_list(self, context: dict):
        with_github_actions = not not path_get(context, "with.github.actions")
        if with_github_actions:
            return [*CPP_EXTENSIONS, "proj_flow.ext.github"]
        return CPP_EXTENSIONS


cxx = api.get_project_type("cxx")



# Copyright (c) 2025 Marcin Zdun
# This code is licensed under MIT license (see LICENSE for details)

"""
The **proj_flow.ext.github.publishing** provides utilities for ``github
publish`` command.
"""


import hashlib
import io
import os
import re
import typing
import zipfile

from proj_flow.api import env, release


def _safe_regex(value: str) -> str:
    for esc in "\\.+*?()[]":
        value = value.replace(esc, f"\\{esc}")
    return value


def build_regex(project: release.Project):
    regexPre = _safe_regex(project.package_prefix)
    regexPost = _safe_regex(project.package_suffix)
    regex = f"^{regexPre}(.*){regexPost}$"
    return re.compile(regex)


def gather_artifacts(directory: str, matcher: re.Pattern):
    names: list[str] = []
    if os.path.isdir(directory):
        for _, dirnames, filenames in os.walk(directory):
            dirnames[:] = []
            names = [name for name in filenames if matcher.match(name)]
    else:
        next_directory = f"{directory}-dir"
        os.makedirs(next_directory, exist_ok=True)
        with zipfile.ZipFile(directory) as zip:
            names = [name for name in zip.namelist() if matcher.match(name)]
            for name in names:
                zip.extract(name, path=next_directory)
        directory = next_directory

    return directory, names


def _hash(filename: str) -> str:
    sha = hashlib.sha256()
    with open(filename, "rb") as data:
        for block in iter(lambda: data.read(io.DEFAULT_BUFFER_SIZE), b""):
            sha.update(block)
    return sha.hexdigest()


def checksums(rt: env.Runtime, directory: str, names: typing.List[str], outname: str):
    rt.print("sha256sum", "-b")
    if not rt.dry_run:
        with open(os.path.join(directory, outname), "w") as output:
            for name in names:
                digest = _hash(os.path.join(directory, name))
                print(f"{digest} *{name}", file=output)
    names.append(outname)



# Copyright (c) 2025 Marcin Zdun
# This code is licensed under MIT license (see LICENSE for details)

"""
The **proj_flow.api.ctx** provides tools for preparing the project template.
"""

import datetime
import inspect
import os
from dataclasses import dataclass
from typing import Any, Callable, Dict, Iterable, List, Optional, Union, cast

from proj_flow.base import cmd
from proj_flow.base import inspect as _inspect

package_root = os.path.dirname(os.path.dirname(__file__))
template_dir = "template"


StrOrBool = Union[str, bool]
Values = Union[StrOrBool, List[str]]
SettingsType = Dict[str, StrOrBool]


@dataclass
class Setting:
    json_key: str
    prompt: str = ""
    value: Union[Values, Callable[[], Values], Callable[[SettingsType], Values]] = ""
    fix: Optional[str] = None
    force_fix: bool = False
    project: Optional[str] = None

    def calc_value(self, previous: SettingsType):
        if callable(self.value):
            kwargs = {}

            params = inspect.signature(self.value).parameters
            if "settings" in params:
                kwargs["settings"] = previous

            return cast(_inspect.Function, self.value)(**kwargs)

        return self.value


def register_init_setting(*settings: Setting, is_hidden=False, project: Optional[str]):
    if project is not None:
        for setting in settings:
            setting.project = project
    (hidden if is_hidden else defaults).extend(settings)


def register_common_init_setting(*setting: Setting, is_hidden=False):
    (hidden if is_hidden else defaults).extend(setting)


def register_switch(key: str, prompt: str, enabled: bool, project: Optional[str]):
    switches.append(Setting(key, prompt, value=enabled, project=project))


def register_common_switch(key: str, prompt: str, enabled: bool):
    switches.append(Setting(key, prompt, value=enabled))


def register_internal(key: str, value: Any):
    internals[key] = value


def _git_config(name: str):
    def wrap():
        proc = cmd.run("git", "config", name, capture_output=True)
        if proc is None or proc.returncode != 0:
            return ""
        return proc.stdout.strip()

    return wrap


def move_to_front(preferred: str, values: Iterable[Optional[str]]):
    result: List[str] = []

    has_preferred = False
    for value in values:
        if value is None:
            continue

        if value == preferred:
            has_preferred = True
        else:
            result.append(value)

    if has_preferred:
        result.insert(0, preferred)

    return result


def _list_ext():
    return move_to_front(".cpp", sorted(_fileext.keys()))


def _as_mustache(basename: str):
    name, ext = os.path.splitext(basename)
    return ext == ".mustache" and name or None


def _enum_licenses():
    root = os.path.abspath(os.path.join(package_root, template_dir, "licenses"))
    for _, dirnames, filenames in os.walk(root):
        dirnames[:] = []
        iter = map(_as_mustache, filenames)
        return move_to_front("MIT", iter)
    return []


def _filter(
    src: Callable[[SettingsType], StrOrBool],
    flt: Callable[[StrOrBool], StrOrBool],
):
    def impl(settings: SettingsType):
        return flt(src(settings))

    return impl


def _get_nothing(_: SettingsType) -> StrOrBool:
    return ""


def _map(internal_key: str):
    def impl(key: StrOrBool) -> StrOrBool:
        mapped = internals.get(internal_key)
        if not isinstance(mapped, dict):
            return ""
        return mapped.get(key, "")

    return impl


def _get_key(key: str):
    def impl(settings: SettingsType):
        return settings.get(key, "")

    return impl


def _build_fixup(settings: SettingsType, fixup: str):
    fixup_refs = fixup.split("{")
    result = fixup_refs.pop(0)
    for ref_expr in fixup_refs:
        ref_text, verbose = ref_expr.split("}")
        filter_path = ref_text.split("$")

        ref = filter_path.pop(0)
        code = _get_nothing if ref == "" else _get_key(ref)

        for filter_name in filter_path:
            flt = _filters.get(filter_name)
            if flt is None:
                if filter_name.startswith("map:"):
                    internal_key = filter_name[4:]
                    code = _filter(code, _map(internal_key))
                continue
            code = _filter(code, flt)

        value = code(settings)

        if result:
            result = f"{result}{value}"
        else:
            result = value

        if verbose:
            result = f"{result}{verbose}"

    return result


def _fixed(fixup: str):
    def wrap(settings: SettingsType):
        return _build_fixup(settings, fixup)

    return wrap


internals: Dict[str, Any] = {}

switches: List[Setting] = []

defaults: List[Setting] = [
    Setting(
        "PROJECT.NAME",
        "Project name",
        lambda: os.path.basename(os.path.abspath("")) or "?project-name?",
    ),
    Setting("PROJECT.DESCRIPTION", "Project description"),
    Setting(
        "PROJECT.EMAIL",
        "Valid email, e.g. for CODE_OF_CONDUCT",
        _git_config("user.email"),
    ),
    Setting(
        "COPY.YEAR",
        "Year for copyright notices",
        str(datetime.date.today().year),
    ),
    Setting(
        "COPY.HOLDER",
        "Holder of the copyright",
        _git_config("user.name"),
    ),
    Setting("COPY.LICENSE", "License", _enum_licenses),
    Setting(
        "INCLUDE_PREFIX",
        'Prefix for includes (as in #include "{PREFIX}/version.hpp")',
        _fixed("{PROJECT.NAME}"),
        project="cxx",
    ),
    Setting(
        "NAME_PREFIX",
        "CMake variable name prefix",
        _fixed("{PROJECT.NAME$safe$upper}"),
        project="cxx",
    ),
    Setting(
        "NAMESPACE",
        "C++ namespace for the project",
        _fixed("{PROJECT.NAME$safe}"),
        project="cxx",
    ),
    Setting("EXT", "Extension for code files", _list_ext, project="cxx"),
    Setting("SRCDIR", "Directory for code files", "src"),
    Setting(
        "INCLUDEDIR",
        "Directory for include files",
        "include",
        "{INCLUDEDIR}/{INCLUDE_PREFIX}",
        force_fix=True,
        project="cxx",
    ),
]

hidden: List[Setting] = [
    Setting("EXT.cxx", fix="{EXT}", project="cxx"),
    Setting("EXT.hxx", fix="{EXT.cxx$header}", project="cxx"),
]

_fileext = {".cc": ".hh", ".cxx": ".hxx", ".cpp": ".hpp"}


_filters: Dict[str, Callable[[StrOrBool], StrOrBool]] = {
    "safe": lambda value: str(value).replace("-", "_"),
    "upper": lambda value: str(value).upper(),
    "lower": lambda value: str(value).lower(),
    "header": lambda cxx_ext: _fileext.get(str(cxx_ext), ".hpp"),
}



# Copyright (c) 2026 Marcin Zdun
# This code is licensed under MIT license (see LICENSE for details)

import typing
from pathlib import Path

import chevron

from proj_flow.api import arg, env
from proj_flow.ext.webidl.base.config import (
    CMakeDirs,
    TemplateConfig,
    load_package_template,
)
from proj_flow.ext.webidl.cli.updater import update_file_if_needed


@arg.command("webidl", "depfile")
def depfile(
    config_path: typing.Annotated[
        str, arg.Argument(help="Input configuration", names=["--cfg"], meta="json-file")
    ],
    binary_dir: typing.Annotated[
        str,
        arg.Argument(
            help="Project binary dir", names=["--binary-dir"], meta="project-binary-dir"
        ),
    ],
    output_path: typing.Annotated[
        str | None,
        arg.Argument(
            help="Output path", names=["--out"], meta="dependency-file", opt=True
        ),
    ],
    rt: env.Runtime,
):
    """Write the dependency file for CMake generators to use"""

    cmake_dirs = CMakeDirs.from_config_path(
        Path(config_path), project_binary_dir=Path(binary_dir)
    )
    global_ctx = cmake_dirs.my_defines()

    config_filename = Path(config_path).absolute()
    config = TemplateConfig.load_config(
        config_filename, global_ctx, cmake_dirs.project_source_dir
    )

    output = (
        Path(output_path)
        if output_path
        else cmake_dirs.dependency_from_config(Path(config_path))
    )

    flat_deps: dict[str, set[str]] = {}

    for rule in config.rules:
        for outname, inputs in rule.get_dependencies(None, config_filename).items():
            try:
                flat_deps[outname].update(inputs)
            except KeyError:
                flat_deps[outname] = set(inputs)
    reverse: dict[str, set[str]] = {}
    for key, value in flat_deps.items():
        val = ";".join(sorted(value))
        try:
            reverse[val].update(key)
        except KeyError:
            reverse[val] = {key}

    rules: list[dict] = []
    for deps, tgts in reverse.items():
        dependency = deps.split(";")
        target = [{"sep": " \\\n", "pathname": pathname} for pathname in tgts]
        target[0]["sep"] = ""
        rules.append({"dependency": dependency, "target": target})

    text = load_package_template("depfile")
    update_file_if_needed(
        output,
        chevron.render(text, {"rule": rules}),
        f"Generating {output.as_posix()}",
    )



# Copyright (c) 2025 Marcin Zdun
# This code is licensed under MIT license (see LICENSE for details)

"""
The **proj_flow.base.inspect** provides unified access to function and class
metadata used in argument parser, command and step metadata.
"""

import inspect
import typing
from dataclasses import dataclass


class Function(typing.Protocol):
    """Replacement for :py:func:`callable` as universal argument type."""

    __name__: str

    def __call__(self, **kwarg) -> typing.Any: ...


@dataclass
class Argument:
    """Extracted argument type"""

    #: Name of the argument
    name: str

    #: The type of the argument. If type was :py:class:`typing.Annotated`, then
    #: this is the first argument to that annotation.
    type: type

    #: The list of metadata attached to the :py:class:`typing.Annotated`, if
    #: there was any.
    metadata: typing.List[typing.Any]


def signature(call: Function) -> typing.Generator[Argument, None, None]:
    """
    Extract the arguments from the function and produces list of
    :class:`Argument` objects.

    :param call: Function to analyse.
    :returns: List of the parameters, in order they appeared in the definition.
    """

    signature = inspect.signature(call)

    for param_name, param in signature.parameters.items():
        annotation = param.annotation

        if typing.get_origin(annotation) is not typing.Annotated:
            yield Argument(name=param_name, type=annotation, metadata=[])
            continue

        anno_args = typing.get_args(annotation)
        if len(anno_args) == 0:
            yield Argument(name=param_name, type=type(None), metadata=[])
            continue

        origin = anno_args[0]
        metadata = list(anno_args[1:])

        yield Argument(name=param_name, type=origin, metadata=metadata)


def type_name(t: type) -> str:
    """
    Converts a type to simplified string representing that type.

    :param t: Type to stringify.
    :returns: String representing the type.
    """

    if t == type(None):
        return "None"

    if type(t) == type:
        return t.__name__

    origin = typing.get_origin(t)

    if origin is typing.Union:
        args = typing.get_args(t)
        names: typing.List[str] = []
        for arg in args:
            name = type_name(arg)
            names.append(name)
        return " | ".join(names)

    if origin is None:
        return "?"

    if origin in [list, type, dict]:
        args = typing.get_args(t)
        arg_list = ", ".join(type_name(arg) for arg in args)
        return f"{origin.__name__}[{arg_list}]"

    args = typing.get_args(t)
    arg_list = ", ".join(type_name(arg) for arg in args)
    if arg_list:
        return f"{origin.__name__}[{arg_list}]"

    return typing.cast(str, origin.__name__)



# Copyright (c) 2026 Marcin Zdun
# This code is licensed under MIT license (see LICENSE for details)

import os
import stat
import subprocess
import sys
from typing import Callable

from proj_flow.ext.test_runner.driver import test
from proj_flow.ext.test_runner.utils.archives import locate_unpack

_file_cache = {}
_rw_mask = stat.S_IWRITE | stat.S_IWGRP | stat.S_IWOTH
_ro_mask = 0o777 ^ _rw_mask


def _touch(test: test.Test, args: list[str]):
    filename = test.path(args[0])
    os.makedirs(os.path.dirname(filename), exist_ok=True)
    with open(filename, "wb") as f:
        if len(args) > 1:
            f.write(args[1].encode("UTF-8"))


def _make_RO(test: test.Test, args: list[str]):
    filename = test.path(args[0])
    mode = os.stat(filename).st_mode
    _file_cache[filename] = mode
    os.chmod(filename, mode & _ro_mask)


def _make_RW(test: test.Test, args: list[str]):
    filename = test.path(args[0])
    try:
        mode = _file_cache[filename]
    except KeyError:
        mode = os.stat(filename).st_mode | _rw_mask
    os.chmod(filename, mode)


def _unpack(test: test.Test, args: list[str]):
    archive = args[0]
    dst = args[1]
    unpack = locate_unpack(archive)[0]
    unpack(archive, test.path(dst))


def _cat(test: test.Test, args: list[str]):
    filename = args[0]
    with open(test.path(filename)) as f:
        sys.stdout.write(f.read())


def _shell(test: test.Test, args: list[str]):
    print("shell!!!")
    print("target:", test.current_env.target if test.current_env is not None else "?")
    subprocess.call("pwsh" if os.name == "nt" else "bash", shell=True, cwd=test.cwd)


HANDLERS: dict[str, tuple[int, Callable[["test.Test", list[str]], None]]] = {
    "mkdirs": (1, lambda test, args: test.makedirs(args[0])),
    "cd": (1, lambda test, args: test.chdir(args[0])),
    "rm": (1, lambda test, args: test.rmtree(args[0])),
    "touch": (1, _touch),
    "unpack": (2, _unpack),
    "store": (2, lambda test, args: test.store_output(args[0], args[1:])),
    "ro": (1, _make_RO),
    "cp": (2, lambda test, args: test.cp(args[0], args[1])),
    "rw": (1, _make_RW),
    "ls": (1, lambda test, args: test.ls(args[0])),
    "cat": (1, _cat),
    "shell": (0, _shell),
}



# Copyright (c) 2025 Marcin Zdun
# This code is licensed under MIT license (see LICENSE for details)

"""
The **proj_flow.ext.cplusplus.cmake** provides ``"CMake"``, ``"Build"``,
``"Pack"`` and ``"Test"`` steps, as well as CMake-specific initialization
context.
"""

from . import parser, project, steps

__all__ = ["parser", "project", "steps"]



# Copyright (c) 2025 Marcin Zdun
# This code is licensed under MIT license (see LICENSE for details)

"""
The **proj_flow.ext.python** defines steps and commands for pyproject packages.
"""

from . import steps, version

__all__ = ["steps", "version"]



# Copyright (c) 2025 Marcin Zdun
# This code is licensed under MIT license (see LICENSE for details)

"""
The **proj_flow.ext.cplusplus** defines steps and commands for C++ projects.
"""

from . import cmake, conan

__all__ = ["cmake", "conan"]



# Copyright (c) 2025 Marcin Zdun
# This code is licensed under MIT license (see LICENSE for details)

import json
import os
from platform import machine
from typing import cast

VAR_NAME = "CXX_FLOW_VERSION"
VARIABLE = ""
IS_FLOW_APP = False

RUNNER = cast(dict, json.loads(os.environ["RUNNER_CONTEXT"]))
RUNNER_OS = RUNNER.get("os", os.name)
RUNNER_ARCH = RUNNER.get("arch", machine() or "unk-arch")

try:
    with open(".flow/flow.py", encoding="UTF-8") as flow_py:
        for line in flow_py:
            if not line.startswith(VAR_NAME):
                continue

            var_list = line.split("=", 1)
            if len(var_list) != 2:
                continue

            var = var_list[0].strip()
            if var != VAR_NAME:
                continue

            val = var_list[1].strip()
            start = val[:1]
            if start not in "\"'":
                continue

            VARIABLE = val[1:].split(start, 1)[0].strip()
            break

    IS_FLOW_APP = True

except FileNotFoundError:
    pass

VENV_CACHE_KEY = f"venv-{RUNNER_OS}-{RUNNER_ARCH}-{VARIABLE}"

with open(os.environ["GITHUB_OUTPUT"], "a", encoding="UTF-8") as out:
    print(f"version={VARIABLE}", file=out)
    print(f"venv-cache-key={VENV_CACHE_KEY}", file=out)
    print(f"is-flow-app={json.dumps(IS_FLOW_APP)}", file=out)



# Copyright (c) 2025 Marcin Zdun
# This code is licensed under MIT license (see LICENSE for details)

"""
The **proj_flow.ext.cplusplus.cmake.steps** defines steps for configuring,
building and archiving.
"""

import os
from pathlib import Path
from typing import Dict, List, cast

from proj_flow import api
from proj_flow.api import env, step
from proj_flow.base.__cmake_version__ import CMAKE_VERSION


class CMakeBase(api.step.Step):
    _name: str
    _runs_after: List[str] = []
    _runs_before: List[str] = []

    @property
    def name(self):
        """:meta private:"""
        return self._name

    @property
    def runs_after(self):
        """:meta private:"""
        return self._runs_after

    @property
    def runs_before(self):
        """:meta private:"""
        return self._runs_before

    def __init__(
        self, name: str, runs_after: List[str] = [], runs_before: List[str] = []
    ):
        super().__init__()
        self._name = name
        self._runs_after = runs_after
        self._runs_before = runs_before

    def is_active(self, config: api.env.Config, rt: api.env.Runtime) -> bool:
        return os.path.isfile("CMakeLists.txt") and os.path.isfile("CMakePresets.json")

    def platform_dependencies(self):
        return [f"cmake>={CMAKE_VERSION}"]

    def dep_with_tool(self, tool: str):
        """:meta private:"""
        return [f"cmake>={CMAKE_VERSION}", f"{tool}>={CMAKE_VERSION}"]


@step.register
class CMakeConfig(CMakeBase):
    """Configures the project using ``preset`` config."""

    def __init__(self):
        super().__init__(name="CMake")

    def is_active(self, config: env.Config, rt: env.Runtime) -> bool:
        return os.path.isfile("CMakeLists.txt") and os.path.isfile("CMakePresets.json")

    def directories_to_remove(self, config: env.Config) -> List[Path]:
        return [config.build_dir]

    def run(self, config: env.Config, rt: env.Runtime) -> int:
        cmake_vars = cast(Dict[str, str], rt._cfg.get("cmake", {}).get("vars", {}))
        defines: List[str] = []
        for var in cmake_vars:
            value = cmake_vars[var]

            is_flag = value.startswith("?")
            if is_flag:
                value = value[1:]

            if value.startswith("config:"):
                value = value[len("config:") :]
                value = config.get_path(value)
            elif value.startswith("runtime:"):
                value = value[len("runtime:") :]
                value = getattr(rt, value, None)

            if is_flag:
                value = "ON" if value else "OFF"

            defines.append(f"-D{var}={value}")

        return rt.cmd(
            "cmake",
            "--preset",
            f"{config.preset}-{config.build_generator}",
            *defines,
        )


@step.register()
class CMakeBuild(CMakeBase):
    """Builds the project using ``preset`` config."""

    def __init__(self):
        super().__init__(name="Build", runs_after=["CMake"])

    def run(self, config: env.Config, rt: env.Runtime) -> int:
        return rt.cmd("cmake", "--build", "--preset", config.preset, "--parallel")


@step.register
class CMakeTest(CMakeBase):
    """Runs tests in the project using ``preset`` config."""

    def __init__(self):
        super().__init__(name="Test", runs_after=["Build"])

    def platform_dependencies(self):
        return self.dep_with_tool("ctest")

    def run(self, config: env.Config, rt: env.Runtime) -> int:
        return rt.cmd("ctest", "--preset", config.preset)


@step.register
class PackStep(CMakeBase):
    """
    Packs archives and installers from ``cpack_generator`` config, using
    ``preset`` config.
    """

    def __init__(self):
        super().__init__(name="Pack", runs_after=["Build"])

    def platform_dependencies(self):
        return self.dep_with_tool("cpack")

    def is_active(self, config: env.Config, rt: env.Runtime) -> bool:
        return (
            super().is_active(config, rt)
            and len(config.items.get("cpack_generator", [])) > 0
        )

    def run(self, config: env.Config, rt: env.Runtime) -> int:
        generators = ";".join(config.items.get("cpack_generator", []))
        return rt.cmd("cpack", "--preset", config.preset, "-G", generators)



# Copyright (c) 2024 Marcin Zdun
# This code is licensed under MIT license (see LICENSE for details)

"""
The **proj_flow.base.registry** allows building extension points, with ability
to register the plugins with a decorator.
"""

import typing
from collections import OrderedDict

T = typing.TypeVar("T")
K = typing.TypeVar("K")


class Registry(typing.Generic[T]):
    """
    Provides simple registry with the decorator attached to it, which
    implements extension point for plugin system. An extension point is a
    value created from this generic, with extendable interface type as
    the generic argument.

    Known decorators
    ................

    :data:`proj_flow.api.release.project_suites`
        :Argument: :class:`ProjectSuite <proj_flow.api.release.ProjectSuite>`
        :Used by: :func:`src.proj_flow.log.release.add_release`

        Project version reader and updater, package file name builder.

    :data:`proj_flow.log.release.version_updaters`
        :Argument: :class:`VersionUpdaters <proj_flow.api.release.VersionUpdaters>`
        :Used by: :func:`src.proj_flow.log.release.add_release`

        Additional version updaters, for instance, path to schema reference on
        GitHub.

    :data:`proj_flow.log.rich_text.api.changelog_generators`
        :Argument: :class:`ChangelogGenerator <proj_flow.log.rich_text.api.ChangelogGenerator>`
        :Used by: :func:`src.proj_flow.ext.github.cli.release`

        Changelog note generator used in CHANGELOG file. Not to confuse with
        generator, which may be used internally by Hosting.add_release.

    Example
    .......

    .. code-block:: python

        class Animal(ABC)
            @abstractmethod
            def speak(self): ...

        animals = Registry[Animal]("Animal")

        def speak_all():
            for animal in animals.get():
                animal.speak()

        @animals.add
        class Dog(Animal):
            def speak(self):
                print("woof!")
    """

    name: str
    container: typing.List[T]

    def __init__(self, name: str):
        self.name = name
        self.container = []
        _debug_copies.append(self)

    def add(self, cls: typing.Type[T]):
        obj: T = cls()
        self.container.append(obj)
        return cls

    def get(self):
        return self.container

    def find(
        self, filter: typing.Callable[[T], K]
    ) -> typing.Tuple[typing.Optional[T], typing.Optional[K]]:
        for item in self.container:
            candidate = filter(item)
            if candidate is not None:
                return item, candidate
        return None, None

    def first(self) -> typing.Optional[T]:
        try:
            return next(self.container.__iter__())
        except StopIteration:
            return None


_debug_copies: typing.List[Registry] = []


def quoted(s: str) -> str:
    if '"' in s:
        return "'{}'".format(s.replace("\\", r"\\").replace("'", r"\'"))
    if "'" in s or " " in s:
        return '"{}"'.format(s)
    return s


def verbose_info():
    for registry in _debug_copies:
        for item in registry.container:
            full_name = f"{item.__module__}.{item.__class__.__name__}"

            kw = OrderedDict()

            if hasattr(item, "name"):
                kw["name"] = quoted(getattr(item, "name"))
            elif hasattr(item, "__name__"):
                kw["name"] = quoted(getattr(item, "__name__"))

            if hasattr(item, "id"):
                kw["id"] = quoted(getattr(item, "id"))

            items = ", ".join([f"{key}={value}" for key, value in kw.items()])
            if len(items) > 0:
                items = f" ({items})"
            print(f"-- {registry.name}: adding `{full_name}`{items}")



# Copyright (c) 2025 Marcin Zdun
# This code is licensed under MIT license (see LICENSE for details)

"""
The **proj_flow.ext.github** provides GitHub support through CLI and GitHub
hosting plugin.
"""

from . import cli, hosting, switches

__all__ = ["cli", "hosting", "switches"]



# Copyright (c) 2026 Marcin Zdun
# This code is licensed under MIT license (see LICENSE for details)

from abc import ABC
from typing import Any

from pywebidl2 import expr

from proj_flow.base.registry import Registry
from proj_flow.ext.webidl.model.ast import MergedDefinitions


class WebIDLVisitor(ABC):
    def on_definitions(self, definitions: MergedDefinitions): ...


class WebIDLCustomExtAttr(ABC):
    def on_attribute(
        self, name: str, ext_attrs: expr.ExtendedAttribute, result: dict[str, Any]
    ): ...


webidl_visitors = Registry[WebIDLVisitor]("WebIDLVisitor")



# Copyright (c) 2025 Marcin Zdun
# This code is licensed under MIT license (see LICENSE for details)

"""
The **proj_flow.ext.sign.api** defines an extension point for per-platform
sign tools.
"""

import base64
import json
import os
from abc import ABC, abstractmethod
from typing import List, NamedTuple, Optional

from proj_flow import base
from proj_flow.api import env

ENV_KEY = "SIGN_TOKEN"


class Key(NamedTuple):
    token: str
    secret: bytes


def _get_key_from_contents(key: str, rt: env.Runtime):
    try:
        obj = json.loads(key)
    except json.decoder.JSONDecodeError:
        rt.message("sign: the signature is not a valid JSON document")
        return None

    if not isinstance(obj, dict):
        rt.message("sign: the signature is missing required fields")
        return None

    token = obj.get("token")
    secret = obj.get("secret")
    if not isinstance(token, str) or not isinstance(secret, str):
        rt.message("sign: the signature is missing required fields")
        return None

    return Key(
        base64.b64decode(token).decode("UTF-8"),
        base64.b64decode(secret),
    )


def get_key(rt: env.Runtime) -> Optional[Key]:
    rt.message(f"sign: trying ${ENV_KEY}")
    env = os.environ.get(ENV_KEY)
    if env:
        key = _get_key_from_contents(env, rt)
        if key is not None:
            return key
    local_signature = os.path.join(".", "signature.key")
    home_signature = os.path.join(os.path.expanduser("~"), "signature.key")
    for filename in [local_signature, home_signature]:
        rt.message(f"sign: trying {filename}")
        if os.path.isfile(filename):
            with open(filename, encoding="UTF-8") as file:
                result = file.read().strip()
                key = _get_key_from_contents(result, rt)
                if key is not None:
                    return key

    rt.message("sign: no key set up")

    return None


class SigningTool(ABC):
    @abstractmethod
    def is_active(self, config: env.Config, rt: env.Runtime): ...

    @abstractmethod
    def sign(self, config: env.Config, rt: env.Runtime, files: List[str]): ...

    @abstractmethod
    def is_signable(self, filename: str, as_package: bool): ...


signing_tool = base.registry.Registry[SigningTool]("SigningTool")



# Copyright (c) 2025 Marcin Zdun
# This code is licensed under MIT license (see LICENSE for details)

"""
The **proj_flow.minimal.list** implements ``./flow list`` command.
"""

import os
import re
import sys
from typing import Annotated, Dict, Iterable, List, Set, cast

from proj_flow import cli
from proj_flow.api import arg, env, step
from proj_flow.base import matrix

if sys.platform == "win32":
    import ctypes
    import ctypes.wintypes
else:
    import termios


@arg.command("list")
def main(
    builtin: Annotated[bool, arg.FlagArgument(help="Show all builtin commands")],
    alias: Annotated[bool, arg.FlagArgument(help="Show all alias commands")],
    steps: Annotated[bool, arg.FlagArgument(help="Show all run steps")],
    configs: Annotated[bool, arg.FlagArgument(help="Show all known matrix keys")],
    all: Annotated[
        bool, arg.FlagArgument(help="Show builtins, aliases, steps and configs")
    ],
    pipe: Annotated[bool, arg.FlagArgument(help="Do not show additional information")],
    rt: env.Runtime,
    menu: cli.argument.Command,
):
    """List all the commands and/or steps for proj-flow"""

    printed_something = False
    bold = "\033[96m" if rt.use_color else ""
    reset = "\033[0m" if rt.use_color else ""

    if all:
        builtin = True
        alias = True
        steps = True
        configs = True

    if builtin:
        builtin_entries = list(sorted(_walk_menu(menu)))
        if not pipe and len(builtin_entries) > 0:
            print("Builtin commands")
            print("----------------")

        for entry_name, entry_doc in builtin_entries:
            if pipe:
                print(entry_name)
                continue

            name = f"{bold}{entry_name}{reset}"
            if entry_doc:
                print(f"- {name}:", end=" ")
                _write_console_para(
                    " ".join(para.split("\n")) for para in entry_doc.split("\n\n")
                )
            else:
                print(f"- {name}")

        printed_something = True

    if alias:
        aliases = rt.aliases

        if not pipe and len(aliases) > 0:
            if printed_something:
                print()

            print("Known aliases")
            print("-------------")

        for run_alias in aliases:
            if pipe:
                print(run_alias.name)
                continue

            name = f"{bold}{run_alias.name}{reset}"
            print(f"- {name}:", end=" ")
            _write_console_para([", ".join(run_alias.steps)])

        printed_something = True

    if steps:
        rt_steps = cast(List[step.Step], rt.steps)

        if not pipe and len(rt_steps) > 0:
            if printed_something:
                print()

            print("Run steps")
            print("---------")

        some_unused = False
        aliased_steps = set(matrix.flatten([alias.steps for alias in rt.aliases]))

        for rt_step in rt_steps:
            if pipe:
                print(rt_step.name)
                continue

            step_used = rt_step.name in aliased_steps
            if not step_used:
                some_unused = True

            name = f"{bold}{rt_step.name}{reset}"
            if step_used:
                print(f"- {name}")
            else:
                print(f"- {name}*")

        if some_unused:
            _write_console_para(
                [
                    f"*step can only be run by explicitly calling through {bold}run{reset}."
                ]
            )

        printed_something = True

    if configs:
        m, keys = _load_flow_data(rt)
        if pipe:
            for key in keys:
                print(key)
        else:
            if len(keys) > 0:
                if printed_something:
                    print()

                print("Matrix keys")
                print("-----------")

            values: Dict[str, Set[str]] = {}

            for config in m:
                for key in keys:
                    value = config[key]
                    if isinstance(value, bool):
                        value = "ON" if value else "OFF"
                    else:
                        value = str(value)
                    try:
                        values[key].add(value)
                    except KeyError:
                        values[key] = {value}

            empty = set()
            for key in keys:
                value = ", ".join(values.get(key, empty))
                name = f"{bold}{key}{reset}"
                if value:
                    print(f"- {name}:", end=" ")
                    _write_console_para([value])
                else:
                    print(f"- {name}")

        printed_something = True

    if not printed_something and not pipe:
        print(f"Use {bold}--help{reset} to see, which listings are available")


def _iterate_levels(menu: cli.argument.Command, prefix: str):
    yield [(f"{prefix}{cmd.name}", cmd.doc) for cmd in menu.children]
    for cmd in menu.children:
        child_prefix = f"{prefix}{cmd.name} "
        for layer in _iterate_levels(cmd, child_prefix):
            yield layer


def _walk_menu(menu: cli.argument.Command):
    root = menu
    while root.parent is not None:
        root = root.parent

    items: list[tuple[str, str]] = []
    for layer in _iterate_levels(root, ""):
        items.extend(layer)

    return items


def _cursor_pos():
    if sys.platform == "win32":
        old_stdin_mode = ctypes.wintypes.DWORD()
        old_stdout_mode = ctypes.wintypes.DWORD()
        kernel32 = ctypes.windll.kernel32
        kernel32.GetConsoleMode(
            kernel32.GetStdHandle(-10), ctypes.byref(old_stdin_mode)
        )
        kernel32.SetConsoleMode(kernel32.GetStdHandle(-10), 0)
        kernel32.GetConsoleMode(
            kernel32.GetStdHandle(-11), ctypes.byref(old_stdout_mode)
        )
        kernel32.SetConsoleMode(kernel32.GetStdHandle(-11), 7)
    else:
        old_stdin_mode = termios.tcgetattr(sys.stdin)
        _ = termios.tcgetattr(sys.stdin)
        _[3] = _[3] & ~(termios.ECHO | termios.ICANON)
        termios.tcsetattr(sys.stdin, termios.TCSAFLUSH, _)
    try:
        _ = ""
        sys.stdout.write("\x1b[6n")
        sys.stdout.flush()
        while not (_ := _ + sys.stdin.read(1)).endswith("R"):
            pass
        res = re.match(r".*\[(?P<y>\d*);(?P<x>\d*)R", _)
    finally:
        if sys.platform == "win32":
            kernel32.SetConsoleMode(kernel32.GetStdHandle(-10), old_stdin_mode)
            kernel32.SetConsoleMode(kernel32.GetStdHandle(-11), old_stdout_mode)
        else:
            termios.tcsetattr(sys.stdin, termios.TCSAFLUSH, old_stdin_mode)
    if res:
        return (int(res.group("x")), int(res.group("y")))
    return (1, 1)


def _write_console_para(text: Iterable[str]):
    term_width = os.get_terminal_size().columns
    margin = min(_cursor_pos()[0], term_width // 2) - 1
    width = term_width - 1
    pos = margin
    for para in text:
        for word in para.split():
            if not word:
                continue
            word_len = len(word)
            next_pos = pos + word_len + 1

            if next_pos >= width:
                orig = pos
                if orig == margin:
                    pos = margin - word_len
                    print(word, end="")

                print()
                print(" " * margin, end="")

                if orig != margin:
                    pos = margin
                    print(word, end=" ")
            else:
                print(word, end=" ")

            pos += word_len + 1

            continue
            if (word_len + 1) > term_width:
                first_word = pos == margin
                if first_word:
                    print(word, end="")

                print(f"\n{' ' * margin}", end="")
                pos = margin

                if not first_word:
                    print(word, "", end="")
                    pos += word_len + 1
                continue

            print(word, "", end="")
            pos += word_len + 1
        print()


def _load_flow_data(rt: env.Runtime):
    paths = [os.path.join(".flow", "matrix.yml")]
    m, keys = matrix.load_matrix(*paths)

    if rt.no_coverage:
        for conf in m:
            if "coverage" in conf:
                del conf["coverage"]

    return m, keys



# Copyright (c) 2025 Marcin Zdun
# This code is licensed under MIT license (see LICENSE for details)

"""
The **proj_flow.ext.cplusplus.conan** provides the ``"Conan"`` step.
"""

from proj_flow.project.cplusplus import project

project.cxx.register_switch("with.conan", "Use Conan for dependency manager", True)



# Copyright (c) 2025 Marcin Zdun
# This code is licensed under MIT license (see LICENSE for details)

"""
The **proj_flow.ext.sign** provides the ``"Sign"`` and ``"SignPackages"``
steps.
"""

import fnmatch
import os
import sys
from abc import abstractmethod
from typing import List, cast

from proj_flow.api import env, init, step

from . import api, win32


def should_exclude(filename: str, exclude: List[str], config_os: str):
    basename = os.path.splitext(filename)[0] if config_os == "windows" else filename

    for pattern in exclude:
        if fnmatch.fnmatch(basename, pattern):
            return True

    return False


class SignBase(step.Step):
    _name: str
    _runs_after: List[str] = []
    _runs_before: List[str] = []

    _active_tools: List[api.SigningTool] = []

    @property
    def name(self):
        return self._name

    @property
    def runs_after(self):
        return self._runs_after

    @property
    def runs_before(self):
        return self._runs_before

    def __init__(
        self, name: str, runs_after: List[str] = [], runs_before: List[str] = []
    ):
        super().__init__()
        self._name = name
        self._runs_after = runs_after
        self._runs_before = runs_before

    def is_active(self, config: env.Config, rt: env.Runtime) -> bool:
        self._active_tools = [
            tool for tool in api.signing_tool.get() if tool.is_active(config, rt)
        ]
        return len(self._active_tools) > 0

    @abstractmethod
    def get_files(
        self, tool: api.SigningTool, config: env.Config, rt: env.Runtime
    ) -> List[str]: ...

    def run(self, config: env.Config, rt: env.Runtime) -> int:
        for tool in self._active_tools:
            files = [
                file.replace(os.sep, "/") for file in self.get_files(tool, config, rt)
            ]

            if len(files) == 0:
                continue

            result = tool.sign(config, rt, files)
            if result:
                return result

        return 0


@step.register
class SignFiles(SignBase):
    """*(Windows)* Signs executable files in build directory"""

    def __init__(self):
        super().__init__(name="Sign", runs_after=["Build"], runs_before=["Pack"])

    def get_files(
        self, tool: api.SigningTool, config: env.Config, rt: env.Runtime
    ) -> List[str]:
        cfg = cast(dict, rt._cfg.get("sign", {}))
        roots = cfg.get("directories", ["bin", "lib", "libexec", "share"])
        exclude = cfg.get("exclude", ["*-test"])

        result: List[str] = []
        build_dir = config.build_dir
        for root in roots:
            for curr_dir, _, filenames in os.walk(os.path.join(build_dir, root)):
                for filename in filenames:
                    if should_exclude(filename, exclude, config.os):
                        continue

                    full_path = os.path.join(curr_dir, filename)
                    if tool.is_signable(full_path, as_package=False):
                        result.append(full_path)
        return result


@step.register
class SignMsi(SignBase):
    """*(Windows)* Signs MSI installers in build directory"""

    def __init__(self):
        super().__init__(
            name="SignPackages",
            runs_after=["Pack"],
            runs_before=["StorePackages", "Store"],
        )

    def is_active(self, config: env.Config, rt: env.Runtime) -> bool:
        return super().is_active(config, rt) and "WIX" in config.items.get(
            "cpack_generator", []
        )

    def get_files(
        self, tool: api.SigningTool, config: env.Config, rt: env.Runtime
    ) -> List[str]:
        result: List[str] = []
        pkg_dir = os.path.join(config.build_dir, "packages")
        for curr_dir, dirnames, filenames in os.walk(pkg_dir):
            dirnames[:] = []
            for filename in filenames:
                full_path = os.path.join(curr_dir, filename)
                if tool.is_signable(full_path, as_package=True):
                    result.append(full_path)

        return result


class SignInit(init.InitStep):
    def postprocess(self, rt: env.Runtime, context: dict):
        with open(".gitignore", "ab") as ignoref:
            ignoref.write("\n/signature.key\n".encode("UTF-8"))


init.register_init_step(SignInit())



# Copyright (c) 2026 Marcin Zdun
# This code is licensed under MIT license (see LICENSE for details)

from typing import cast

from proj_flow.api import arg
from proj_flow.base import inspect as _inspect


@arg.command("webidl")
def webidl():
    """Configure and generate WebIDL-based interfaces"""



# Copyright (c) 2026 Marcin Zdun
# This file is licensed under MIT license (see LICENSE for details)

from pathlib import Path


def update_file_if_needed(dst: Path, contents: str, message: str | None = None):
    prev = ""
    if dst.exists():
        prev = dst.read_bytes().decode("utf-8")
    if prev == contents:
        return False

    if message:
        print("-- WebIDL:", "\n-- WebIDL: ".join(message.split("\n")))

    dst.parent.mkdir(parents=True, exist_ok=True)
    dst.write_bytes(contents.encode("utf-8"))

    return True



# Copyright (c) 2026 Marcin Zdun
# This code is licensed under MIT license (see LICENSE for details)


import copy
import json
from dataclasses import dataclass, field
from pathlib import Path
from typing import Any, Callable, cast

from pywebidl2 import expr

package_root = Path(__file__).parent.parent

TR = list[str] | str | None


@dataclass
class TypeReplacement:
    name: str
    module_or_include: str | None = None

    @staticmethod
    def load_config(lang: str | None, types: Path | None):
        data = {}
        if lang:
            path = package_root / "data" / "types" / f"{lang}.json"
            with path.open(encoding="utf-8") as lang_file:
                data.update(cast(dict, json.load(lang_file)))

        if types:
            with types.open(encoding="utf-8") as lang_file:
                all_types_data = cast(dict[str, dict], json.load(lang_file))
                if lang:
                    data.update(all_types_data.get(lang, {}))
                else:
                    data.update(all_types_data.get("", {}))

        result: dict[str, TypeReplacement] = {}
        for key, value in data.items():
            if value is None:
                result[key] = TypeReplacement(name=key)

            elif isinstance(value, str):
                result[key] = TypeReplacement(name=value)

            elif isinstance(value, list) and len(value) == 2:
                mod, name = value
                if isinstance(name, str) and (mod is None or isinstance(mod, str)):
                    result[key] = TypeReplacement(name=name, module_or_include=mod)

        return result


def visit_flag_extended_attribute(
    name: str, bag: dict[str, Any], attributes: dict[str, expr.ExtendedAttribute]
):
    bag[name] = attributes.get(name) != None


def visit_string_extended_attribute(
    name: str, bag: dict[str, Any], attributes: dict[str, expr.ExtendedAttribute]
):
    bag[name] = None
    try:
        attr = attributes[name]
    except KeyError:
        return

    if attr.rhs:
        value = cast(str, attr.rhs.value)
        if value.startswith('"') and value.endswith('"') and len(value) > 1:
            value = value[1:-1]
        bag[name] = value


def flag_extended_attribute(name: str):
    def impl(bag: dict[str, Any], attributes: dict[str, expr.ExtendedAttribute]):
        return visit_flag_extended_attribute(name, bag, attributes)

    return impl


def string_extended_attribute(name: str):
    def impl(bag: dict[str, Any], attributes: dict[str, expr.ExtendedAttribute]):
        return visit_string_extended_attribute(name, bag, attributes)

    return impl


ATTR_TYPE = {
    "bool": flag_extended_attribute,
    "str": string_extended_attribute,
}

ExtAttrsVisitor = Callable[[dict[str, Any], dict[str, expr.ExtendedAttribute]], None]


@dataclass
class ExtAttrsContextBuilder:
    visitors: dict[str, ExtAttrsVisitor] = field(default_factory=dict)
    applies_to_type: bool = False

    def add_flag(self, name: str):
        self.visitors[name] = flag_extended_attribute(name)

    def add_string(self, name: str):
        self.visitors[name] = string_extended_attribute(name)

    def visit(self, bag: dict[str, Any], attributes: dict[str, expr.ExtendedAttribute]):
        for visitor in self.visitors.values():
            visitor(bag, attributes)

    def merge(self, rhs: "ExtAttrsContextBuilder"):
        left = copy.deepcopy(self.visitors)
        right = copy.deepcopy(rhs.visitors)
        left.update(right)
        return ExtAttrsContextBuilder(
            visitors=left,
            applies_to_type=self.applies_to_type,
        )


class BuilderBuilder:
    tgt: ExtAttrsContextBuilder

    def __init__(self, applies_to_type: bool = False):
        self.tgt = ExtAttrsContextBuilder(applies_to_type=applies_to_type)

    def flag(self, name: str):
        self.tgt.add_flag(name)
        return self

    def string(self, name: str):
        self.tgt.add_string(name)
        return self

    def build(self):
        return self.tgt


@dataclass
class ExtAttrsContextBuilders:
    enum: ExtAttrsContextBuilder = field(default_factory=ExtAttrsContextBuilder)
    interface: ExtAttrsContextBuilder = field(default_factory=ExtAttrsContextBuilder)
    attribute: ExtAttrsContextBuilder = field(default_factory=ExtAttrsContextBuilder)
    operation: ExtAttrsContextBuilder = field(default_factory=ExtAttrsContextBuilder)
    return_type: ExtAttrsContextBuilder = field(default_factory=ExtAttrsContextBuilder)
    argument: ExtAttrsContextBuilder = field(default_factory=ExtAttrsContextBuilder)
    type: ExtAttrsContextBuilder = field(default_factory=ExtAttrsContextBuilder)

    def from_idl(
        self,
        selector: Callable[["ExtAttrsContextBuilders"], ExtAttrsContextBuilder],
        ext_attrs: list[expr.ExtendedAttribute],
    ):
        result: dict[str, Any] = {}
        attributes = {attr.name: attr for attr in ext_attrs}
        builder = selector(self)
        builder.visit(result, attributes)
        if builder.applies_to_type:
            self.type.visit(result, attributes)
        return result

    def merge(self, rhs: "ExtAttrsContextBuilders"):
        return ExtAttrsContextBuilders(
            enum=self.enum.merge(rhs.enum),
            interface=self.interface.merge(rhs.interface),
            attribute=self.attribute.merge(rhs.attribute),
            operation=self.operation.merge(rhs.operation),
            return_type=self.return_type.merge(rhs.return_type),
            argument=self.argument.merge(rhs.argument),
            type=self.type.merge(rhs.type),
        )

    @staticmethod
    def from_config(data: dict):
        ext_attrs = cast(dict[str, dict[str, str]], data.get("ext_attrs", {}))
        result = ExtAttrsContextBuilders()
        targets = [
            ("enum", result.enum),
            ("interface", result.interface),
            ("attribute", result.attribute),
            ("operation", result.operation),
            ("return_type", result.return_type),
            ("argument", result.argument),
        ]
        for key, target in targets:
            group = ext_attrs.get(key, {})
            for name, type_name in group.items():
                ctor = ATTR_TYPE.get(type_name)
                if not ctor:
                    message = f"unknown user-defined extended attribute `{type_name}' when reading {key}.{name}"
                    raise RuntimeError(message)
                target.visitors[name] = ctor(name)
        return result

    def property_fixup(self, type: dict[str, Any], parent: dict[str, Any]):
        for key in self.type.visitors:
            if key in parent:
                value = parent[key]
                del parent[key]
                if isinstance(value, bool):
                    type[key] = cast(bool, type.get(key, False)) or value
                    continue

                if isinstance(value, str):
                    type[key] = cast(str, type.get(key, False)) or value
                    continue

    @staticmethod
    def builtin():
        return ExtAttrsContextBuilders(
            attribute=BuilderBuilder().string("default").build(),
            operation=BuilderBuilder()
            .flag("in")
            .flag("out")
            .flag("mutable")
            .flag("throws")
            .build(),
            argument=BuilderBuilder()
            .flag("defaulted")
            .string("default")
            .flag("in")
            .flag("out")
            .build(),
            type=BuilderBuilder()
            .flag("unique")
            .flag("span")
            .flag("nullable")
            .string("key")
            .build(),
        )



# Copyright (c) 2025 Marcin Zdun
# This code is licensed under MIT license (see LICENSE for details)

"""
The **proj_flow.ext.cplusplus.cmake.__version__** contains only the value of
expected version of CMake.
"""


CMAKE_VERSION = "3.28"



# Copyright (c) 2025 Marcin Zdun
# This code is licensed under MIT license (see LICENSE for details)

"""
The **proj_flow.api.env** provides interaction with project environment:

- :class:`FlowConfig` loads the project config in ``.flow/config.yml``.
- :class:`Runtime` provides means to run tools and print messages, while
  respecting ``--dry-run``, ``--silent`` and ``--verbose``.
- :class:`Config` represents single run configuration.
"""

import argparse
import importlib
import json
import os
import re
import shlex
import shutil
import subprocess
import sys
from dataclasses import dataclass
from enum import Enum
from pathlib import Path
from typing import Any, Callable, Dict, List, Optional, Union, cast

from proj_flow.api import ctx
from proj_flow.base import plugins, uname
from proj_flow.base.cmake_presets import get_binary_dirs

platform = uname.uname()[0]


_flow_config_default_compiler: Optional[Dict[str, str]] = None


class CMakeBinaryDirsCache:
    _dirs: dict[str, Path] | None = None
    _loaded: bool = False

    @property
    def dirs(self):
        if not self._loaded:
            self._loaded = True
            self._dirs = get_binary_dirs()

        return self._dirs

    def binary_dir(self, key: str):
        return (self.dirs or {}).get(key)


CMAKE_BINARY_DIRS = CMakeBinaryDirsCache()


def default_compiler():
    try:
        return os.environ["DEV_CXX"]
    except KeyError:
        pass

    flow_config_default_compiler = _flow_config_default_compiler or {}

    try:
        return flow_config_default_compiler[platform]
    except KeyError:
        print(
            f"-- KeyError: {platform} in {flow_config_default_compiler}",
            file=sys.stderr,
        )
        return "?"
    except TypeError:
        print(f"-- TypeError: internal: flow config not ready yet", file=sys.stderr)
        return "?"


class Printer:
    @staticmethod
    def hide(arg: str, secrets: List[str]):
        for secret in secrets:
            arg = arg.replace(secret, "?" * max(15, len(secret)))
        return arg

    @staticmethod
    def print_arg(arg: str, secrets: List[str], raw: bool):
        color = ""
        arg = Printer.hide(arg, secrets)
        if arg[:1] == "-":
            color = "\033[2;37m"
        if not raw:
            arg = shlex.join([arg])
        if color == "" and arg[:1] in ["'", '"']:
            color = "\033[2;34m"
        if color == "":
            return arg
        return f"{color}{arg}\033[m"

    @staticmethod
    def print_cmd(*args: str, use_color: bool = True, secrets: List[str], raw: bool):
        cmd = args[0] if raw else shlex.join([args[0]])
        if not use_color:
            if raw:
                print(
                    cmd,
                    *(Printer.hide(arg, secrets) for arg in args[1:]),
                    file=sys.stderr,
                )
            else:
                print(
                    cmd,
                    shlex.join(Printer.hide(arg, secrets) for arg in args[1:]),
                    file=sys.stderr,
                )
            return

        printed = " ".join([Printer.print_arg(arg, secrets, raw) for arg in args[1:]])
        print(f"\033[33m{cmd}\033[m {printed}", file=sys.stderr)


@dataclass
class RunAlias:
    name: str
    doc: str
    steps: List[str]

    @staticmethod
    def from_json(name: str, alias: Union[dict, list]):
        if isinstance(alias, dict):
            doc: str = alias.get("doc", "")
            steps: List[str] = alias.get("steps", [])
        else:
            doc = ""
            steps = alias
        if not doc:
            doc = f'Shortcut for "run -s {",".join(steps)}"'

        return RunAlias(name, doc, steps)


def _merge_dicts(dst: dict, src: dict):
    for key in src:
        if key not in dst:
            dst[key] = src[key]
            continue

        src_val = src[key]
        dst_val = dst[key]

        if isinstance(src_val, dict):
            if isinstance(dst_val, dict):
                _merge_dicts(dst_val, src_val)
                continue

        dst[key] = src_val


def _flatten_dict(dst: ctx.SettingsType, defaults: Dict[str, Any], prefix=""):
    for key, val in defaults.items():
        this_key = f"{prefix}{key}"

        if isinstance(val, dict):
            _flatten_dict(dst, val, f"{this_key}.")
            continue

        if val is None:
            dst[this_key] = ""
            continue

        if isinstance(val, bool):
            dst[this_key] = val
            continue

        dst[this_key] = str(val)


def _merge(cfg: dict, defaults: ctx.SettingsType, path: Path):
    config = plugins.load_data(path)
    if not isinstance(config, dict):
        return

    stored_defaults = config.get("defaults", {})
    try:
        del config["defaults"]
    except KeyError:
        pass

    _merge_dicts(cfg, config)

    if isinstance(stored_defaults, dict):
        _flatten_dict(defaults, stored_defaults)


def load_extensions(extensions: List[str]):
    for extension in extensions:
        try:
            importlib.import_module(extension)
        except ImportError as ex:
            print(
                f"-- error: loading module `{extension}` resulted in import error: {ex}",
                file=sys.stderr,
            )
        except BaseException as ex:
            print(
                f"-- error: there was an error while loading module `{extension}`: {ex}",
                file=sys.stderr,
            )


class FlowConfig:
    _cfg: dict
    steps: list = []
    aliases: List[RunAlias] = []
    root: Path

    def __init__(self, cfg: Optional["FlowConfig"] = None, root: Path = Path()):
        if cfg is not None:
            self._cfg = cfg._cfg
            self.steps = cfg.steps
            self.aliases = cfg.aliases
            self.root = cfg.root
        else:
            self.root = root.absolute()
            defaults: ctx.SettingsType = {}
            dest: dict = {}

            _merge(
                dest,
                defaults,
                Path("~").expanduser() / ".config" / "proj-flow.json",
            )
            _merge(dest, defaults, self.root / ".flow" / "config.json")

            self._cfg = dest
            self._cfg["defaults"] = defaults

            self._propagate_compilers()
            self._load_extensions()

    def _propagate_compilers(self):
        global _flow_config_default_compiler
        _flow_config_default_compiler = self.compiler_os_default

    def _load_extensions(self):
        extensions = cast(List[str], self._cfg.get("extensions", []))
        extensions.insert(0, "proj_flow.minimal")

        local_extensions = (self.root / ".flow" / "extensions").resolve()
        if local_extensions.is_dir():
            sys.path.insert(0, local_extensions.as_posix())

        load_extensions(extensions)

    @property
    def entry(self) -> Dict[str, dict]:
        return self._cfg.get("entry", {})

    @property
    def compiler(self) -> Dict[str, dict]:
        return self._cfg.get("compiler", {})

    @property
    def compiler_names(self) -> Dict[str, List[str]]:
        return self.compiler.get("names", {})

    @property
    def compiler_os_default(self) -> Dict[str, str]:
        return self.compiler.get("os-default", {})

    @property
    def lts_list(self) -> Dict[str, List[str]]:
        return self._cfg.get("lts", {})

    @property
    def postproc(self) -> dict:
        return self._cfg.get("postproc", {})

    @property
    def postproc_exclude(self) -> List[dict]:
        return self.postproc.get("exclude", [])

    @property
    def postproc_include(self) -> List[dict]:
        return self.postproc.get("include", [])

    @property
    def shortcuts(self) -> Dict[str, dict]:
        return self._cfg.get("shortcuts", {})


def _mkdir(dirname: str):
    os.makedirs(dirname, exist_ok=True)


def _ls(dirname: str, shallow=True):
    result = []
    for root, dirnames, filenames in os.walk(dirname):
        if shallow:
            dirnames[:] = []

        result.extend(
            os.path.relpath(os.path.join(root, filename), start=dirname)
            for filename in filenames
        )
    return result


def _cp(src: str, dst: str) -> int:
    try:
        dst = os.path.abspath(dst)
        if os.path.isdir(src):
            _mkdir(dst)
            shutil.copytree(src, dst, dirs_exist_ok=True, symlinks=True)
        else:
            if not os.path.isdir(dst):
                _mkdir(os.path.dirname(dst))
            shutil.copy(src, dst, follow_symlinks=False)
        return 0
    except FileNotFoundError as err:
        print(err, file=sys.stderr)
        return 1


class Msg(Enum):
    """Message level for Runtime.message"""

    DEBUG = 0
    """Print only, when verbose is set"""

    STATUS = 2
    """Print when silent if not set"""

    ALWAYS = 3
    """Print always"""


MSG_GUARD: Dict[Msg, Callable[["Runtime"], bool]] = {
    Msg.DEBUG: lambda rt: rt.verbose,
    Msg.STATUS: lambda rt: not rt.silent,
    Msg.ALWAYS: lambda rt: True,
}


class Runtime(FlowConfig):
    dry_run: bool
    silent: bool
    verbose: bool
    official: bool
    no_coverage: bool
    use_color: bool
    only_host: bool
    platform: str
    secrets: List[str] = []

    def __init__(
        self, argsOrRuntime: Union[argparse.Namespace, "Runtime"], cfg: FlowConfig
    ):
        super().__init__(cfg=cfg)

        if isinstance(argsOrRuntime, argparse.Namespace):
            args = argsOrRuntime
            self.dry_run = getattr(args, "dry_run", False)
            self.silent = getattr(args, "silent", False)
            self.verbose = getattr(args, "verbose", False)
            self.official = getattr(args, "official", False)
            self.use_color = True
            self.no_coverage = False
            self.platform = platform

            if "NO_COVERAGE" in os.environ:
                self.no_coverage = True

            if "RELEASE" in os.environ and "GITHUB_ACTIONS" in os.environ:
                self.official = not not json.loads(os.environ["RELEASE"])

            self.only_host = not (self.dry_run or self.official)
        else:
            rt = argsOrRuntime
            self.dry_run = rt.dry_run
            self.silent = rt.silent
            self.verbose = rt.verbose
            self.official = rt.official
            self.no_coverage = rt.no_coverage
            self.use_color = rt.use_color
            self.only_host = rt.only_host
            self.platform = rt.platform
            self.secrets = [*rt.secrets]

    def message(self, *args: str, level=Msg.DEBUG, **kwargs):
        if not MSG_GUARD[level](self):
            return

        print("--", *args, **kwargs, file=sys.stderr)

    def fatal(self, *args: str, **kwargs):
        print("-- FATAL:", *args, **kwargs, file=sys.stderr)
        sys.exit(1)

    def print(self, *args: str, raw=False):
        if not self.silent:
            Printer.print_cmd(
                *args, use_color=self.use_color, secrets=self.secrets, raw=raw
            )

    def cmd(self, *args: str):
        self.print(*args)
        if self.dry_run:
            return 0

        result = subprocess.run(args)
        if result.returncode != 0:
            print(
                f"proj-flow: error: {args[0]} ended in failure, exiting",
                file=sys.stderr,
            )
            raise SystemExit(1)
        return 0

    def capture(self, *args: str, silent=False):
        if not silent:
            self.print(*args)
        return subprocess.run(args, shell=False, encoding="UTF-8", capture_output=True)

    def mkdirs(self, dirname: str):
        self.print("mkdir", "-p", dirname)

        if self.dry_run:
            return 0

        os.makedirs(dirname, exist_ok=True)
        return 0

    def cp(self, src: str, dst: str, regex: Optional[str] = None):
        args = ["cp"]
        if os.path.isdir(src):
            args.append("-r")
        self.print(*args, src, dst)

        if self.dry_run:
            return 0

        if regex is None:
            return _cp(src, dst)

        files = _ls(src)
        files = (name for name in files if re.match(regex, name))
        for name in files:
            result = _cp(os.path.join(src, name), os.path.join(dst, name))
            if result:
                return result
        return 0


@dataclass
class Config:
    items: dict
    keys: List[str]

    def get_path(self, key: str, val: Any = None):
        path = key.split(".")
        ctx = self.items
        for step in path:
            if not isinstance(ctx, dict):
                return val
            child = ctx.get(step)
            if child is None:
                return val
            ctx = child
        return cast(Any, ctx)

    @property
    def os(self) -> str:
        return self.items.get("os", "")

    @property
    def compiler(self) -> List[str]:
        return self.items.get("compiler", [])

    @property
    def build_type(self) -> str:
        return self.items.get("build_type", "")

    @property
    def build_name(self) -> str:
        return self.items.get("build_name", "")

    @property
    def build_dir(self) -> Path:
        cmake_binary_dir = CMAKE_BINARY_DIRS.binary_dir(
            f"{self.preset}-{self.build_generator}"
        )
        if cmake_binary_dir:
            return cmake_binary_dir
        return Path("build") / self.preset

    @property
    def preset(self) -> str:
        return self.items.get("preset", "")

    @property
    def build_generator(self) -> str:
        return self.items.get("build_generator", "")



# Copyright (c) 2025 Marcin Zdun
# This code is licensed under MIT license (see LICENSE for details)

"""
The **proj_flow.ext.python.version** provides project suite plugin.
"""

import os
import re
from typing import NamedTuple, Optional, cast

import toml

from proj_flow.api import env, release


class QuickProjectInfo(NamedTuple):
    name: Optional[str] = None
    path: Optional[str] = None
    pattern: Optional[str] = None


@release.project_suites.add
class ProjectSuite(release.ProjectSuite):
    def get_project(self, rt: env.Runtime) -> Optional[release.Project]:
        name, path, pattern = self._pyproject_hatch(rt)
        if name is None or path is None:
            return None

        try:
            with open(rt.root / path, encoding="UTF-8") as in_file:
                text = in_file.read()
        except FileNotFoundError:
            return None

        if pattern is not None:
            candidate = self._check(text, pattern)
            return release.Project(name, candidate) if candidate else None

        for varname in ["__version__", "VERSION"]:
            for end in ['"', "'"]:
                pattern = rf"{varname}\s*=\s*{end}v?(?P<version>[^{end}]+){end}"
                candidate = self._check(text, pattern)
                if candidate:
                    return release.Project(name, candidate)

        return None

    def _check(self, text: str, pattern: str):
        m = re.search(f"^{pattern}", text, flags=re.MULTILINE)
        if not m:
            return None
        version, start = m.group("version"), m.start("version")
        core = re.split(r"([0-9]+\.[0-9]+\.[0-9]+)", version, maxsplit=1)[1]
        stability = version[len(core) :]

        return release.Version(
            release.Arg(core, start), release.Arg(stability, start + len(core))
        )

    def get_version_file_path(self, rt: env.Runtime) -> Optional[str]:
        _, path, _ = self._pyproject_hatch(rt)
        return path

    def _pyproject_hatch(self, rt: env.Runtime):
        try:
            data = toml.load(rt.root / "pyproject.toml")
            project = cast(dict, data.get("project", {}))
            hatch = cast(dict, data.get("tool", {}).get("hatch", {}))
            wheels = cast(
                dict,
                hatch.get("build", {})
                .get("targets", {})
                .get("wheel", {})
                .get("packages", []),
            )

            name = cast(str | None, project.get("name"))
            if len(wheels) > 0:
                first_wheel = wheels[0].split("/")[-1]
                if first_wheel:
                    name = first_wheel

            dynamic = cast(list[str], project.get("dynamic", []))
            if "version" in dynamic:
                version_dict = cast(dict[str, str | None], hatch.get("version", {}))
                return QuickProjectInfo(
                    name=name,
                    path=version_dict.get("path"),
                    pattern=version_dict.get("pattern"),
                )
            return QuickProjectInfo(
                name=name,
                path="pyproject.toml",
            )
        except FileNotFoundError:
            return QuickProjectInfo()



# Copyright (c) 2024 Marcin Zdun
# This code is licensed under MIT license (see LICENSE for details)

"""
The **proj_flow.log.rich_text.api** defines base types for CHANGELOG.md
manipulation.
"""

import abc
import os
import re
from typing import Dict, List, Type

from proj_flow import base
from proj_flow.api import env
from proj_flow.log import commit, msg


class FileUpdate(msg.ReleaseMessage):
    commit_date: str

    def __init__(self, setup: commit.LogSetup, commit_date: str):
        super().__init__(setup)
        self.commit_date = commit_date


class ChangelogGenerator(abc.ABC):
    @property
    def filename(self) -> str:
        return f"CHANGELOG{self.ext}"

    @property
    @abc.abstractmethod
    def formatter(self) -> Type[FileUpdate]: ...

    @property
    @abc.abstractmethod
    def ext(self) -> str: ...

    @property
    @abc.abstractmethod
    def split_re(self) -> str: ...

    @property
    def intro_head(self) -> str:
        return "Changelog"

    @property
    def intro_text(self) -> str:
        return "All notable changes to this project will be documented in this file."

    @abc.abstractmethod
    def main_heading(self, caption: str) -> List[str]: ...

    @abc.abstractmethod
    def para(self, text: str) -> List[str]: ...

    def intro(self):
        lines = self.main_heading(self.intro_head)
        lines.extend(self.para(self.intro_text))
        return "\n".join(lines)

    def format_changelog(
        self, log: commit.ChangeLog, setup: commit.LogSetup, rt: env.Runtime
    ):
        formatter = self.formatter(
            setup, commit.read_tag_date(setup.curr_tag or "HEAD", rt)
        )
        return formatter.format_changelog(log)

    def create_changelog(
        self,
        tags: List[str],
        git: commit.Git,
        links: commit.Hosting,
        rt: env.Runtime,
        scope_fix: Dict[str, str] = {},
        take_all: bool = False,
    ):
        prev_tag = None
        entire_log: List[str] = []
        for curr_tag in tags:
            setup = commit.LogSetup(
                links, prev_tag, curr_tag, scope_fix=scope_fix, take_all=take_all
            )
            log, _ = git.get_log(setup)
            text = self.format_changelog(log, setup, rt)
            prev_tag = curr_tag

            entire_log.append(text + "\n")

        entire_log.append(self.intro())

        filename = rt.root / f"CHANGELOG{self.ext}"
        with open(filename, "wb") as f:
            for text in reversed(entire_log):
                f.write(text.encode("UTF-8"))

    def update_changelog(
        self,
        log: commit.ChangeLog,
        setup: commit.LogSetup,
        rt: env.Runtime,
    ):
        formatter = self.formatter(
            setup, commit.read_tag_date(setup.curr_tag or "HEAD", rt)
        )
        text = formatter.format_changelog(log)
        path = rt.root / self.filename

        try:
            with open(path, encoding="UTF-8") as f:
                changelog = f.read()
        except FileNotFoundError:
            changelog = self.intro()

        split = re.split(
            pattern=f"({self.split_re})",
            flags=re.MULTILINE,
            string=changelog,
            maxsplit=1,
        )
        new_text = split[0] + "\n" + text
        if len(split) > 1:
            new_text += "".join(split[1:])

        with open(path, "wb") as f:
            f.write(new_text.encode("UTF-8"))


changelog_generators = base.registry.Registry[ChangelogGenerator]("ChangelogGenerator")



# Copyright (c) 2025 Marcin Zdun
# This code is licensed under MIT license (see LICENSE for details)

"""
The **proj_flow.project.cplusplus.cmake_presets** check build directories for CMake step.
"""

import json
import os
import platform
from dataclasses import dataclass
from enum import Enum
from pathlib import Path
from typing import Callable, TypeVar, cast


class MacroType(Enum):
    PRESET_NAME = 1
    GENERATOR = 2
    FILE_DIR = 3


CMAKE_SOURCE_DIR = Path(".").resolve()
CMAKE_HOST_SYSTEM_NAME = platform.uname().system

MACROS: dict[str, Path | str | MacroType] = {
    "${sourceDir}": CMAKE_SOURCE_DIR,
    "${sourceParentDir}": CMAKE_SOURCE_DIR.parent,
    "${sourceDirName}": CMAKE_SOURCE_DIR.name,
    "${presetName}": MacroType.PRESET_NAME,
    "${generator}": MacroType.GENERATOR,
    "${hostSystemName}": CMAKE_HOST_SYSTEM_NAME,
    "${fileDir}": MacroType.FILE_DIR,
    "${dollar}": "$",
    "${pathListSep}": os.pathsep,
}


@dataclass
class Preset:
    name: str
    binary_dir: str | None
    build_type: str | None
    generator: str | None
    inherits: list[str]
    file_dir: Path

    def expand(self):
        if not self.binary_dir:
            return None
        chunks = self.binary_dir.split("${")
        parts: list[str | Path | MacroType] = [chunks[0]]
        for chunk in chunks[1:]:
            var_suffix = chunk.split("}", 1)
            if len(var_suffix) != 2:
                return None
            key = f"${{{var_suffix[0]}}}"
            value = MACROS.get(key, "")
            if isinstance(value, MacroType):
                if value == MacroType.PRESET_NAME:
                    value = self.name
                elif value == MacroType.GENERATOR:
                    value = self.generator or ""
                elif value == MacroType.FILE_DIR:
                    value = self.file_dir
            elif isinstance(value, Path):
                if value.is_absolute():
                    parts = []
                value = str(value)
            parts.append(value)
            parts.append(var_suffix[1])

        return Path("".join(str(part) for part in parts))


class Presets:
    def __init__(self):
        self.seen = cast(set[Path], set())
        self.result = cast(dict[str, Preset], {})

    def visit_file(self, filename: Path):
        dirname = self.__parent_from_file(filename)
        if dirname is None:
            return self.result

        try:
            includes, presets = Presets.__load_file(filename)
            self.__visit_includes(dirname, includes)
            for preset in presets:
                self.__visit_preset(dirname, preset)
            for preset in self.result.values():
                self.__merge_preset(preset)
        except FileNotFoundError:
            return None

        return self.result

    def __parent_from_file(self, filename: Path):
        key = filename.resolve()
        if key in self.seen:
            return None
        self.seen.add(key)
        return key.parent

    def __visit_includes(self, dirname: Path, includes: list[str]):
        for include in includes:
            self.visit_file(dirname / include)

    def __visit_preset(self, dirname: Path, preset: dict):
        name = cast(str | None, preset.get("name"))
        if name is None:
            return
        binary_dir = cast(str | None, preset.get("binaryDir"))
        generator = cast(str | None, preset.get("generator"))
        inherits = cast(list[str], preset.get("inherits", []))
        cache = cast(dict, preset.get("cacheVariables", {}))
        build_type = cast(str | None, cache.get("CMAKE_BUILD_TYPE"))
        self.result[name] = Preset(
            name=name,
            binary_dir=binary_dir,
            build_type=build_type,
            generator=generator,
            inherits=inherits,
            file_dir=dirname,
        )

    def __merge_preset(self, preset: Preset):
        inherits: list[str] = []
        inherits.extend(preset.inherits)
        while inherits:
            parent = inherits[0]
            inherits = inherits[1:]
            if parent not in self.result:
                continue
            parent_preset = self.result[parent]
            inherits.extend(parent_preset.inherits)

            if parent_preset.binary_dir and not preset.binary_dir:
                preset.binary_dir = parent_preset.binary_dir
            if parent_preset.generator and not preset.generator:
                preset.generator = parent_preset.generator
            if parent_preset.build_type and not preset.build_type:
                preset.build_type = parent_preset.build_type

    @staticmethod
    def __load_file(filename: Path):
        with open(filename, encoding="UTF-8") as f:
            data = json.load(f)
        includes = cast(list[str], data.get("include", []))
        presets = cast(list[dict], data.get("configurePresets", []))

        return (includes, presets)


def binary_dir_from_preset(preset: Preset, cwd: Path):
    path = preset.expand()
    if not path:
        return None
    if path.is_relative_to(cwd):
        path = path.relative_to(cwd)
    return path


def get_binary_dirs():
    return visit_presets(binary_dir_from_preset)


T = TypeVar("T")


def visit_presets(mapper: Callable[[Preset, Path], T | None]):
    presets = Presets().visit_file(Path("CMakePresets.json"))
    if presets is None:
        return None

    cwd = Path.cwd()
    result: dict[str, T] = {}
    for preset in presets.values():
        mapped = mapper(preset, cwd)
        if mapped is None:
            continue
        result[preset.name] = mapped

    return result



# Copyright (c) 2024 Marcin Zdun
# This code is licensed under MIT license (see LICENSE for details)

"""
The **proj_flow.log.rich_text** allows to operate on CHANGELOG.md.
"""

from . import api, markdown, re_structured_text

__all__ = ["api", "markdown", "re_structured_text"]



# Copyright (c) 2025 Marcin Zdun
# This code is licensed under MIT license (see LICENSE for details)

"""
The **proj_flow.ext.cplusplus.cmake.version** provides project suite plugin.
"""

from typing import NamedTuple, Optional

from proj_flow.api import env, release
from proj_flow.ext.cplusplus.cmake.parser import get_project


class QuickProjectInfo(NamedTuple):
    name: Optional[str] = None
    path: Optional[str] = None
    pattern: Optional[str] = None


@release.project_suites.add
class ProjectSuite(release.ProjectSuite):
    def get_project(self, rt: env.Runtime) -> Optional[release.Project]:
        project = get_project(rt.root)
        if project is None:
            return None
        return release.Project(
            project.name.value, release.Version(project.version, project.stability)
        )

    def get_version_file_path(self, rt: env.Runtime) -> Optional[str]:
        return "CMakeLists.txt"



# Copyright (c) 2026 Marcin Zdun
# This code is licensed under MIT license (see LICENSE for details)

import os
import tarfile
import zipfile
from typing import Callable


def _untar(src, dst):
    with tarfile.open(src) as TAR:

        def is_within_directory(directory, target):
            abs_directory = os.path.abspath(directory)
            abs_target = os.path.abspath(target)

            prefix = os.path.commonprefix([abs_directory, abs_target])

            return prefix == abs_directory

        for member in TAR.getmembers():
            member_path = os.path.join(dst, member.name)
            if not is_within_directory(dst, member_path):
                raise Exception(f"Attempted path traversal in Tar file: {member.name}")

        TAR.extractall(dst)


def _unzip(src, dst):
    with zipfile.ZipFile(src) as ZIP:
        ZIP.extractall(dst)


_tar = (_untar, ["tar", "-xf"])

Unpacker = Callable[[str, str], None]
UnpackInfo = tuple[Unpacker, list[str]]

ARCHIVES: dict[str, UnpackInfo] = {
    ".tar": _tar,
    ".tar.gz": _tar,
    ".zip": (_unzip, ["unzip"]),
}


def locate_unpack(archive: str) -> UnpackInfo:
    reminder, ext = os.path.splitext(archive)
    _, mid = os.path.splitext(reminder)
    if mid == ".tar":
        ext = ".tar"
    return ARCHIVES[ext]


del _tar
del _unzip
del _untar



# Copyright (c) 2025 Marcin Zdun
# This code is licensed under MIT license (see LICENSE for details)

"""
The **proj_flow.cli.argument** provides command-line builders and runners,
supporting the functions decorated with :func:`@arg.command()
<proj_flow.api.arg.command>`.
"""

import argparse
import typing
from dataclasses import dataclass, field

from proj_flow import __version__
from proj_flow.api import arg, completers, env, step
from proj_flow.base import inspect as _inspect
from proj_flow.base import registry
from proj_flow.flow import configs


class Completer(typing.Protocol):
    def __call__(self, **kwarg) -> typing.Any: ...


class Action(typing.Protocol):
    completer: Completer


class Subparsers(typing.Protocol):
    parent: "Parser"

    def add_parser(*args, **kwargs) -> "Parser": ...


class Parser(argparse.ArgumentParser):
    flow: env.FlowConfig
    menu: typing.List["Command"]
    shortcuts: typing.Dict[str, typing.List[str]]

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def add_subparsers(self, **kwargs) -> Subparsers:
        return typing.cast(Subparsers, super().add_subparsers(**kwargs))

    def find_and_run_command(self, args: argparse.Namespace):
        commands = {entry.name for entry in self.menu}
        aliases = self.flow.aliases

        rt = env.Runtime(args, self.flow)

        if rt.verbose:
            verbose_info(self.menu)
            step.verbose_info()
            registry.verbose_info()

        if args.command in commands:
            command = _first(lambda command: command.name == args.command, self.menu)
            if command:
                return command.run(args, rt)
        elif args.command in {alias.name for alias in aliases}:
            command = _first(lambda command: command.name == "run", self.menu)
            alias = _first(lambda alias: alias.name == args.command, aliases)
            if command and alias:
                args.cli_steps.append(",".join(alias.steps))
                return command.run(args, rt)

        lines = ["the command arguments are required; known commands:", ""]
        for command in self.menu:
            lines.append(f"  - {command.name}: {command.doc}")
        for alias in aliases:
            lines.append(f"  - {alias.name}: {alias.doc}")

        self.error("\n".join(lines))


def build_argparser(flow_cfg: env.FlowConfig):
    parser = Parser(
        prog="proj-flow",
        description="C++ project maintenance, automated",
        add_help=False,
    )
    parser.add_argument(
        "-h", "--help", action="help", help="Show this help message and exit"
    )
    parser.add_argument(
        "-v",
        "--version",
        action="version",
        default=argparse.SUPPRESS,
        version=f"%(prog)s version {__version__}",
        help="Show proj-flow's version and exit",
    )
    parser.add_argument(
        "-C",
        metavar="dir",
        nargs="?",
        help="Run as if proj-flow was started in <dir> instead of the current "
        "working directory. This directory must exist.",
    ).completer = completers.cd_completer  # type: ignore

    menu = _build_menu(arg.get_commands(), None)
    _argparse_visit_all(parser, flow_cfg, menu.children)
    return parser


def expand_shortcuts(parser: Parser, args: argparse.Namespace):
    args_kwargs = dict(args._get_kwargs())
    shortcuts: typing.Dict[str, typing.List[str]] = parser.shortcuts
    for key in shortcuts:
        try:
            if not args_kwargs[key]:
                continue
            typing.cast(typing.List[str], args.configs).extend(shortcuts[key])
            break
        except KeyError:
            continue


@dataclass
class AdditionalArgument:
    name: str
    ctor: callable  # type: ignore

    def create(self, rt: env.Runtime, args: argparse.Namespace, menu: "Command"):
        if self.ctor == env.Runtime:
            return rt
        if self.ctor == Command:
            return menu
        return self.ctor(rt, args)


@dataclass
class AnnotatedArgument:
    name: str
    argument: arg.Argument

    @property
    def group(self):
        return self.argument.group

    def argparse_visit(self, parser: argparse._ActionsContainer):
        return self.argument.visit(parser, self.name)


AnyArgument = typing.Union[AdditionalArgument, AnnotatedArgument]


@dataclass
class Command:
    name: str
    doc: str
    entry: callable  # type: ignore
    annotated: typing.List[AnnotatedArgument]
    additional: typing.List[AdditionalArgument]
    parent: typing.Optional["Command"]
    children: typing.List["Command"] = field(default_factory=list)

    def argparse_visit(
        self,
        subparsers: Subparsers,
        alias: typing.Optional[str] = None,
        doc: typing.Optional[str] = None,
        level=0,
    ):
        if not doc:
            doc = self.doc
        if not alias:
            alias = self.name

        parser = subparsers.add_parser(
            alias, help=doc.split("\n\n")[0], description=doc, add_help=False
        )

        parent = subparsers.parent
        parser.flow = parent.flow
        parser.shortcuts = parent.shortcuts

        assert parent.flow is not None
        assert parent.shortcuts is not None

        _argparse_runtime_visit(parser)

        has_config = False
        for additional in self.additional:
            if additional.ctor == configs.Configs:
                has_config = True
                break

        if has_config:
            _argparse_config_visit(parser)

        groups: typing.Set[int] = set()
        for arg_index in range(len(self.annotated)):
            annotated = self.annotated[arg_index]
            group = annotated.group
            if group is not None:
                if id(group) in groups:
                    continue
                groups.add(id(group))
                group_container = group.visit(parser)
                for sub_index in range(arg_index, len(self.annotated)):
                    sub = self.annotated[sub_index]
                    if sub.group is group:
                        sub.argparse_visit(group_container)
                continue
            annotated.argparse_visit(parser)

        if len(self.children):
            subparsers = parser.add_subparsers(
                dest=f"command_{level}",
                metavar="command",
                help="Known command name, see below",
            )
            subparsers.parent = parser  # type: ignore

            for entry in self.children:
                entry.argparse_visit(subparsers, level=level + 1)

    def run(self, args: argparse.Namespace, rt: env.Runtime, level=0):
        if level == 0 and rt.only_host:
            rt.only_host = self.name == "run"

        subcommand_name = None

        if len(self.children):
            subcommand_attribute = f"command_{level}"
            if hasattr(args, subcommand_attribute):
                subcommand_name = getattr(args, subcommand_attribute)

        if subcommand_name is not None:
            subcommand = _first(
                lambda command: command.name == subcommand_name, self.children
            )
            if not subcommand:
                rt.fatal(f"cannot find {subcommand_name}")

            return subcommand.run(args, rt, level=level + 1)

        kwargs = {}
        for arg in self.annotated:
            kwargs[arg.name] = getattr(args, arg.name, None)

        for additional in self.additional:
            arg = additional.create(rt, args, self)
            kwargs[additional.name] = arg

        result = self.entry(**kwargs)
        return 0 if result is None else result


T = typing.TypeVar("T")


def _first(
    fltr: typing.Callable[[T], bool], items: typing.Iterable[T]
) -> typing.Optional[T]:
    try:
        return next(filter(fltr, items))
    except StopIteration:
        return None


def _argparse_visit_all(
    parser: Parser, cfg: env.FlowConfig, menu: typing.List["Command"]
):
    shortcut_configs = _build_run_shortcuts(cfg)

    parser.menu = menu
    parser.flow = cfg
    parser.shortcuts = shortcut_configs

    subparsers = parser.add_subparsers(
        dest="command", metavar="command", help="Known command name, see below"
    )

    subparsers.parent = parser

    run: typing.Optional[Command] = None
    for entry in menu:
        entry.argparse_visit(subparsers)
        if entry.name == "run":
            run = entry

    if run is not None and len(cfg.aliases) > 0:
        commands = {entry.name for entry in menu}
        cfg.aliases = [alias for alias in cfg.aliases if alias.name not in commands]
        for alias in cfg.aliases:
            run.argparse_visit(subparsers, alias=alias.name, doc=alias.doc)
    else:
        cfg.aliases = []

    pass


def _build_menu(cmd: arg._Command, parent: typing.Optional[Command]):
    name = cmd.name
    doc = cmd.doc or ""
    entry = cmd.entry or (lambda: 0)

    args = _extract_args(entry)
    additional = [entry for entry in args if isinstance(entry, AdditionalArgument)]
    annotated = [entry for entry in args if isinstance(entry, AnnotatedArgument)]

    current = Command(
        name=name,
        doc=doc,
        entry=entry,
        annotated=annotated,
        additional=additional,
        parent=parent,
    )
    for child in cmd.subs.values():
        current.children.append(_build_menu(child, current))

    return current


def _extract_arg(argument: _inspect.Argument):
    for ctor in [configs.Configs, env.Runtime, Command]:
        if argument.type is ctor:
            return AdditionalArgument(argument.name, ctor)

    metadata: typing.Optional[arg.Argument] = _first(
        lambda meta: isinstance(meta, arg.Argument), argument.metadata
    )

    if metadata is None or argument.type is None:
        return None

    optional = metadata.opt
    if optional is None:
        NoneType = type(None)
        is_union = typing.get_origin(argument.type) is typing.Union
        optional = is_union and NoneType in typing.get_args(argument.type)
    metadata.opt = optional

    return AnnotatedArgument(argument.name, metadata)


def _extract_args(entry: callable):  # type: ignore
    args_with_possible_nones = map(_extract_arg, _inspect.signature(entry))
    args = filter(lambda item: item is not None, args_with_possible_nones)
    return typing.cast(typing.List[AnyArgument], list(args))


def _build_run_shortcuts(cfg):
    shortcut_configs: typing.Dict[str, typing.List[str]] = {}
    args: typing.List[typing.Tuple[str, typing.List[str], bool, bool]] = []

    shortcuts = cfg.shortcuts
    for shortcut_name in sorted(shortcuts.keys()):
        has_os = False
        has_compiler = False
        shortcut = shortcuts[shortcut_name]
        config: typing.List[str] = []
        for key in sorted(shortcut.keys()):
            has_os = has_os or key == "os"
            has_compiler = has_compiler or key == "os"
            value = shortcut[key]
            if isinstance(value, list):
                for v in value:
                    config.append(f"{key}={_shortcut_value(v)}")
            else:
                config.append(f"{key}={_shortcut_value(value)}")
        if len(config) > 0:
            args.append((shortcut_name, config, has_os, has_compiler))

    if len(args):
        os_prefix = f"os={env.platform}"
        compiler_prefix = f"compiler={env.default_compiler()}"

        for shortcut_name, config, has_os, has_compiler in args:
            if not has_compiler:
                config.insert(0, compiler_prefix)
            if not has_os:
                config.insert(0, os_prefix)
            shortcut_configs[shortcut_name] = config

    return shortcut_configs


def _shortcut_value(value) -> str:
    if isinstance(value, bool):
        return "ON" if value else "OFF"
    return str(value)


def _argparse_runtime_visit(parser: Parser):
    parser.add_argument(
        "-h",
        "--help",
        action="help",
        default=argparse.SUPPRESS,
        help="Show this help message and exit",
    )

    parser.add_argument(
        "--dry-run",
        action="store_true",
        required=False,
        help="Print steps and commands, do nothing",
    )

    verbosity = parser.add_mutually_exclusive_group()
    verbosity.add_argument(
        "--silent",
        action="store_true",
        required=False,
        help="Remove most of the output",
    )
    verbosity.add_argument(
        "--verbose",
        action="store_true",
        required=False,
        help="Add even more output",
    )


def _argparse_config_visit(parser: Parser):
    parser.add_argument(
        "-D",
        dest="configs",
        metavar="key=value",
        nargs="*",
        action="store",
        default=[],
        help="Run only builds on matching configs. The key is one of "
        'the keys into "matrix" object in .flow/matrix.yml definition '
        "and the value is one of the possible values for that key. In "
        "case of boolean flags, such as sanitizer, the true value is "
        'one of "true", "on", "yes", "1" and "with-<key>", '
        'i.e. "with-sanitizer" for sanitizer.'
        " "
        "If given key is never used, all values from .flow/matrix.yaml "
        "for that key are used. Otherwise, only values from command "
        "line are used.",
    ).completer = completers.matrix_completer  # type: ignore

    parser.add_argument(
        "--official",
        action="store_true",
        required=False,
        help="Cut matrix to release builds only",
    )

    if len(parser.shortcuts):
        group = parser.add_mutually_exclusive_group()

        for shortcut_name in sorted(parser.shortcuts.keys()):
            config = parser.shortcuts[shortcut_name]
            group.add_argument(
                f"--{shortcut_name}",
                required=False,
                action="store_true",
                help=f'Shortcut for "-D {" ".join(config)}"',
            )


def verbose_info(commands: typing.List[Command], prefix=""):
    for command in commands:
        cli = f"{prefix} {command.name}" if prefix else command.name
        if command.entry is not None:
            print(
                f"-- Command: adding `{cli}` from `{command.entry.__module__}.{command.entry.__name__}(...)`"
            )
        verbose_info(command.children, cli)



# Copyright (c) 2025 Marcin Zdun
# This code is licensed under MIT license (see LICENSE for details)

"""
The **proj_flow.ext.markdown_changelog** registers a
:class:`api.ChangelogGenerator` plugin implemented for markdown changelogs.
"""

from proj_flow.log.rich_text.api import changelog_generators
from proj_flow.log.rich_text.markdown import ChangelogGenerator


@changelog_generators.add
class Plugin(ChangelogGenerator):
    pass
